from typing import Callable, Optional, Union

import numpy as np

from parea.evals.utils import embed
from parea.schemas import Log


def semantic_similarity_factory(
    embd_model: str = "text-embedding-3-small",
    is_azure: Optional[bool] = False,
) -> Callable[[Log], Union[float, None]]:
    def semantic_similarity(log: Log) -> Union[float, None]:
        """Calculates semantic similarity between output and target"""
        output = log.output
        if (target := log.target) is None:
            return None

        output_vector = embed(model=embd_model, input=output, is_azure=is_azure)
        target_vector = embed(model=embd_model, input=target, is_azure=is_azure)
        output_vector = np.array(output_vector)
        target_vector = np.array(target_vector)

        return (np.dot(output_vector, target_vector) / (np.linalg.norm(output_vector) * np.linalg.norm(target_vector)) + 1) / 2

    return semantic_similarity


semantic_similarity_oai_3_small = semantic_similarity_factory()
semantic_similarity_oai_3_large = semantic_similarity_factory(embd_model="text-embedding-3-large")
semantic_similarity_oai_ada_002 = semantic_similarity_factory(embd_model="text-embedding-ada-002")



from openai import AsyncStream, Stream


class OpenAIAsyncStreamWrapper:
    def __init__(self, async_stream: AsyncStream, accumulator, info_from_response, update_accumulator_streaming, final_processing_and_logging):
        self._async_stream = async_stream
        self._final_processing_and_logging = final_processing_and_logging
        self._update_accumulator_streaming = update_accumulator_streaming
        self._accumulator = accumulator
        self._info_from_response = info_from_response

    def __getattr__(self, attr):
        # delegate attribute access to the original async_stream
        return getattr(self._async_stream, attr)

    async def __aiter__(self):
        async for chunk in self._async_stream:
            self._update_accumulator_streaming(self._accumulator, self._info_from_response, chunk)
            yield chunk

        self._final_processing_and_logging(self._accumulator, self._info_from_response)


class OpenAIStreamWrapper:
    def __init__(self, stream: Stream, accumulator, info_from_response, update_accumulator_streaming, final_processing_and_logging):
        self._stream = stream
        self._final_processing_and_logging = final_processing_and_logging
        self._update_accumulator_streaming = update_accumulator_streaming
        self._accumulator = accumulator
        self._info_from_response = info_from_response

    def __getattr__(self, attr):
        # delegate attribute access to the original async_stream
        return getattr(self._stream, attr)

    def __iter__(self):
        for chunk in self._stream:
            self._update_accumulator_streaming(self._accumulator, self._info_from_response, chunk)
            yield chunk

        self._final_processing_and_logging(self._accumulator, self._info_from_response)



from typing import Any, Callable, Optional, Tuple, Union

import asyncio
import contextvars
import datetime
import functools
import inspect
import json
import logging
import os
import traceback

import cohere
from cohere import NonStreamedChatResponse, RerankResponse

from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.helpers import gen_trace_id, is_logging_disabled, timezone_aware_now
from parea.schemas import LLMInputs, ModelParams, TraceLog, UpdateTraceScenario
from parea.utils.trace_utils import execution_order_counters, fill_trace_data, logger_record_log, trace_context, trace_data
from parea.utils.universal_encoder import json_dumps
from parea.wrapper.cohere.helpers import DEFAULT_MODEL, DEFAULT_P, DEFAULT_TEMPERATURE, chat_history_to_messages, get_output, get_usage_stats

logger = logging.getLogger()


class CohereClientWrapper:
    @staticmethod
    def wrap_method(method: Callable, method_name: str) -> Callable:
        """
        Wrap a method with logging functionality based on its type.

        :param method: The original method to be wrapped
        :param method_name: The name of the method for logging purposes
        :return: Wrapped method
        """

        @functools.wraps(method)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            if is_logging_disabled():
                return method(*args, **kwargs)

            trace_id, start_time, context_token = CohereClientWrapper.init_trace(method_name, args, kwargs, method)
            try:
                if method_name == "chat_stream":
                    return CohereClientWrapper._handle_stream(trace_id, start_time, context_token, method, *args, **kwargs)

                result = method(*args, **kwargs)
                if not is_logging_disabled():
                    CohereClientWrapper._fill_llm_config(trace_id, result, **kwargs)
                return result
            except Exception as e:
                logger.error(f"Error occurred in function {method.__name__}, {e}", exc_info=True)
                fill_trace_data(trace_id, {"error": traceback.format_exc()}, UpdateTraceScenario.ERROR)
                raise e
            finally:
                try:
                    if method_name != "chat_stream":
                        CohereClientWrapper.cleanup_trace(trace_id, start_time, context_token)
                except Exception as e:
                    logger.debug(f"Error occurred cleaning up trace for function {method.__name__}, {e}", exc_info=e)

        @functools.wraps(method)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            if is_logging_disabled():
                return await method(*args, **kwargs)

            trace_id, start_time, context_token = CohereClientWrapper.init_trace(method_name, args, kwargs, method)
            try:
                result = await method(*args, **kwargs)
                if not is_logging_disabled():
                    CohereClientWrapper._fill_llm_config(trace_id, result, **kwargs)
                return result
            except Exception as e:
                logger.error(f"Error occurred in function {method.__name__}, {e}")
                fill_trace_data(trace_id, {"error": traceback.format_exc()}, UpdateTraceScenario.ERROR)
                raise e
            finally:
                try:
                    if method_name != "chat_stream":
                        CohereClientWrapper.cleanup_trace(trace_id, start_time, context_token)
                except Exception as e:
                    logger.debug(f"Error occurred cleaning up trace for function {method.__name__}, {e}", exc_info=e)

        if asyncio.iscoroutinefunction(method):
            return async_wrapper
        else:
            return sync_wrapper

    @staticmethod
    def init_trace(func_name, args, kwargs, func) -> Tuple[str, datetime, contextvars.Token]:
        start_time = timezone_aware_now()
        trace_id = gen_trace_id()

        new_trace_context = trace_context.get() + [trace_id]
        token = trace_context.set(new_trace_context)

        try:
            inputs = None
            if kwargs.get("query", None):
                sig = inspect.signature(func)
                parameters = sig.parameters

                inputs = {k: v for k, v in zip(parameters.keys(), args)}
                inputs.update(kwargs)

                # filter out any values which aren't JSON serializable
                for k, v in inputs.items():
                    try:
                        json.dumps(v)
                    except TypeError:
                        try:
                            inputs[k] = json_dumps(v)
                        except (TypeError, AttributeError):
                            # if we can't serialize the value, just convert it to a string
                            inputs[k] = str(v)

            depth = len(new_trace_context) - 1
            root_trace_id = new_trace_context[0]

            # Get the execution order counter for the current root trace
            counters = execution_order_counters.get()
            if root_trace_id not in counters:
                counters[root_trace_id] = 0
            execution_order = counters[root_trace_id]
            counters[root_trace_id] += 1

            trace_data.get()[trace_id] = TraceLog(
                trace_id=trace_id,
                parent_trace_id=root_trace_id,
                root_trace_id=root_trace_id,
                start_timestamp=start_time.isoformat(),
                trace_name="llm-cohere",
                inputs=inputs,
                experiment_uuid=os.environ.get(PAREA_OS_ENV_EXPERIMENT_UUID, None),
                depth=depth,
                execution_order=execution_order,
            )
            parent_trace_id = new_trace_context[-2] if len(new_trace_context) > 1 else None
            if parent_trace_id:
                fill_trace_data(trace_id, {"parent_trace_id": parent_trace_id}, UpdateTraceScenario.CHAIN)
        except Exception as e:
            logger.debug(f"Error occurred initializing trace for function {func_name}, {e}")

        return trace_id, start_time, token

    @staticmethod
    def cleanup_trace(trace_id: str, start_time: datetime, context_token: contextvars.Token, time_to_first_token: Optional[float] = None) -> None:
        end_time = timezone_aware_now()
        trace_data.get()[trace_id].end_timestamp = end_time.isoformat()
        trace_data.get()[trace_id].latency = (end_time - start_time).total_seconds()
        if time_to_first_token:
            trace_data.get()[trace_id].time_to_first_token = time_to_first_token
        logger_record_log(trace_id)
        trace_context.reset(context_token)

    @staticmethod
    def _handle_stream(trace_id: str, start_time: datetime, context_token: contextvars.Token, method: Callable, *args: Any, **kwargs: Any):
        accumulated_text = ""
        final_response: Optional[NonStreamedChatResponse] = None
        time_to_first_token = None

        try:
            for event in method(*args, **kwargs):
                if not time_to_first_token:
                    time_to_first_token = (timezone_aware_now() - start_time).total_seconds()
                if event.event_type == "text-generation":
                    accumulated_text += event.text
                elif event.event_type == "stream-end":
                    final_response = event.response

                yield event
        except Exception as e:
            logger.error(f"Error occurred in function {method.__name__}, {e}")
            fill_trace_data(trace_id, {"error": traceback.format_exc()}, UpdateTraceScenario.ERROR)
            raise e
        finally:
            CohereClientWrapper._fill_llm_config(trace_id, final_response, **kwargs)
            CohereClientWrapper.cleanup_trace(trace_id, start_time, context_token, time_to_first_token)

    @staticmethod
    def _fill_llm_config(trace_id: str, result: Optional[NonStreamedChatResponse | RerankResponse], **kwargs: dict[str, Any]) -> None:
        """
        Fill the LLM configuration data for the given trace.

        :param trace_id: The ID of the current trace
        :param result: The response from the Cohere API
        :param kwargs: Additional keyword arguments
        """
        try:
            model = kwargs.get("model", DEFAULT_MODEL)
            tools = kwargs.get("tools", None)
            configuration = LLMInputs(
                model=model,
                provider="cohere",
                model_params=ModelParams(
                    temp=kwargs.get("temperature", DEFAULT_TEMPERATURE),
                    top_p=kwargs.get("p", DEFAULT_P),
                    frequency_penalty=kwargs.get("frequency_penalty", 0),
                    presence_penalty=kwargs.get("presence_penalty", 0),
                    max_length=kwargs.get("max_tokens"),
                    response_format=kwargs.get("response_format"),
                ),
                messages=chat_history_to_messages(result, **kwargs) if isinstance(result, NonStreamedChatResponse) else None,
                functions=tools,
            )
            prompt_tokens, completion_tokens, cost = get_usage_stats(result, model)
            data = {
                "configuration": configuration,
                "output": get_output(result),
                "input_tokens": prompt_tokens,
                "output_tokens": completion_tokens,
                "total_tokens": prompt_tokens + completion_tokens,
                "cost": cost,
            }
            fill_trace_data(trace_id, data, UpdateTraceScenario.OPENAICONFIG)
        except Exception as e:
            logger.debug(f"Error occurred filling LLM config for trace {trace_id}, {e}", exc_info=True)
            fill_trace_data(trace_id, {"error": traceback.format_exc()}, UpdateTraceScenario.ERROR)

    @staticmethod
    def init(client: Union[cohere.Client, cohere.AsyncClient]) -> None:
        """
        Apply the CohereClientWrapper to the 'chat', 'chat_stream', and 'rerank' methods of the cohere.Client instance.

        :param client: An instance of cohere.Client
        """
        methods_to_wrap = ["chat", "chat_stream", "rerank"]
        for method_name in methods_to_wrap:
            if hasattr(client, method_name):
                original_method = getattr(client, method_name)
                wrapped_method = CohereClientWrapper.wrap_method(original_method, method_name)
                setattr(client, method_name, wrapped_method)



from typing import Callable, Optional

import json

from parea.evals import call_openai
from parea.schemas import Log


def context_has_answer_factory(question_field: Optional[str] = "question", model: Optional[str] = "gpt-3.5-turbo-0125", is_azure: Optional[bool] = False) -> Callable[[Log], bool]:
    """
    This factory creates an evaluation metric which assess whether the given context has the answer to the given question.
    It is useful to measure the performance of a model in a question-answering task by measuring Hit Rate without the need to know the correct answer.

    Args:
        question_field: The key name/field used for the question/query of the user. Defaults to "question".
        model: The model which should be used for grading. Currently, only supports OpenAI chat models. Defaults to "gpt-3.5-turbo-0125".
        is_azure: Whether to use the Azure API. Defaults to False.

    Returns:
        Callable[[Log], bool]: A function that takes a log as input and returns a boolean indicating if the context has the answer to the given question.
    """

    def context_has_answer(log: Log) -> bool:
        question = log.inputs[question_field]
        answer = str(log.output)

        formatted_messages = [
            {
                "role": "user",
                "content": f"""You are given a question and a list of answers. The answers were retrieved from a database which contains the question answer pairs. You need to decide if any of the given answers is the answer to the given question.

Question:
{question}

Answers:
{answer}

Answer in the following JSON format:
{{"thoughts": "<thoughts>", "final_verdict": "<true|false>"}}""",
            }
        ]

        response = call_openai(model=model, temperature=0.0, messages=formatted_messages, response_format={"type": "json_object"}, is_azure=is_azure)
        final_verdict = json.loads(response).get("final_verdict", "").lower()
        return final_verdict == "true"

    return context_has_answer



# type: ignore[attr-defined]
# flake8: noqa

"""
    Parea API SDK

    The Parea SDK allows you to interact with Parea from your product or service.
    To install the official [Python SDK](https://pypi.org/project/parea/),
    run the following command:  ```bash pip install parea ```.
"""
import sys

from parea.api_client import get_version
from parea.cache import InMemoryCache
from parea.client import Parea
from parea.experiment.cli import experiment as _experiment_cli
from parea.experiment.dvc import parea_dvc_initialized
from parea.experiment.experiment import Experiment
from parea.helpers import gen_trace_id, write_trace_logs_to_csv
from parea.parea_logger import parea_logger
from parea.utils.trace_utils import clear_trace_context, get_current_trace_id, get_root_trace_id, trace, trace_insert
from parea.wrapper.openai_raw_api_tracer import aprocess_stream_and_yield, process_stream_and_yield
from parea.wrapper.utils import convert_openai_raw_to_log

version: str = get_version()


def main():
    args = sys.argv[1:]
    if args[0] == "experiment":
        _experiment_cli(args[1:])
    elif args[0] == "dvc-init":
        parea_dvc_initialized(only_check=False)
    else:
        print(f"Unknown command: '{args[0]}'")


if __name__ == "__main__":
    main()



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import json
from enum import Enum

from attrs import define, field, validators

from parea.schemas import EvaluationResult
from parea.schemas.log import EvaluatedLog, LLMInputs


@define
class Completion:
    inference_id: Optional[str] = None
    parent_trace_id: Optional[str] = None
    root_trace_id: Optional[str] = None
    trace_name: Optional[str] = None
    llm_inputs: Optional[Dict[str, Any]] = None
    llm_configuration: LLMInputs = LLMInputs()
    end_user_identifier: Optional[str] = None
    deployment_id: Optional[str] = None
    name: Optional[str] = None
    metadata: Optional[Dict] = None
    tags: Optional[List[str]] = field(factory=list)
    target: Optional[str] = None
    cache: bool = True
    log_omit_inputs: bool = False
    log_omit_outputs: bool = False
    log_omit: bool = False
    log_sample_rate: float = 1.0
    experiment_uuid: Optional[str] = None
    project_uuid: str = "default"


@define
class CompletionResponse:
    inference_id: str
    content: str
    latency: float
    input_tokens: int
    output_tokens: int
    total_tokens: int
    cost: float
    model: str
    provider: str
    cache_hit: bool
    status: str
    start_timestamp: str
    end_timestamp: str
    error: Optional[str] = None


@define
class UseDeployedPrompt:
    deployment_id: str
    llm_inputs: Optional[Dict[str, Any]] = None


@define
class Prompt:
    raw_messages: List[Dict[str, Any]]
    messages: List[Dict[str, Any]]
    inputs: Optional[Dict[str, Any]] = None


@define
class UseDeployedPromptResponse:
    deployment_id: str
    version_number: float
    name: Optional[str] = None
    functions: Optional[List[str]] = None
    function_call: Optional[str] = None
    prompt: Optional[Prompt] = None
    model: Optional[str] = None
    provider: Optional[str] = None
    model_params: Optional[Dict[str, Any]] = None


@define
class FeedbackRequest:
    score: float = field(validator=[validators.ge(0), validators.le(1)])
    trace_id: Optional[str] = None
    name: Optional[str] = None
    target: Optional[str] = None


@define
class TraceLogImage:
    url: str
    caption: Optional[str] = None


@define
class TraceLogCommentSchema:
    comment: str
    user_id: str
    created_at: str


@define
class TraceLogAnnotationSchema:
    created_at: str
    user_id: str
    score: float
    user_email_address: Optional[str] = None
    annotation_name: Optional[str] = None
    value: Optional[str] = None


@define
class TraceLog(EvaluatedLog):
    trace_id: Optional[str] = field(default=None, validator=validators.instance_of(str))
    parent_trace_id: Optional[str] = field(default=None, validator=validators.instance_of(str))
    root_trace_id: Optional[str] = field(default=None, validator=validators.instance_of(str))
    start_timestamp: Optional[str] = field(default=None, validator=validators.instance_of(str))
    organization_id: Optional[str] = None
    project_uuid: Optional[str] = None

    # metrics filled from completion
    error: Optional[str] = None
    status: Optional[str] = None
    deployment_id: Optional[str] = None
    cache_hit: bool = False
    output_for_eval_metrics: Optional[str] = None
    evaluation_metric_names: Optional[List[str]] = field(factory=list)
    apply_eval_frac: float = 1.0
    log_sample_rate: float = 1.0
    feedback_score: Optional[float] = None

    # info filled from decorator
    trace_name: Optional[str] = None
    children: List[str] = field(factory=list)

    # metrics filled from either decorator or completion
    end_timestamp: Optional[str] = None
    end_user_identifier: Optional[str] = None
    session_id: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    tags: Optional[List[str]] = field(factory=list)
    experiment_uuid: Optional[str] = None
    images: Optional[List[TraceLogImage]] = field(factory=list)

    # from UI
    comments: Optional[List[TraceLogCommentSchema]] = None
    annotations: Optional[Dict[int, Dict[str, TraceLogAnnotationSchema]]] = None

    depth: int = 0
    execution_order: int = 0


@define
class TraceLogTree(TraceLog):
    children_logs: Optional[List["TraceLogTree"]] = field(factory=list)


@define
class CacheRequest:
    configuration: LLMInputs = LLMInputs()


@define
class UpdateLog:
    trace_id: str
    field_name_to_value_map: Dict[str, Any]
    root_trace_id: Optional[str] = None


@define
class CreateExperimentRequest:
    name: str
    run_name: str
    metadata: Optional[Dict[str, str]] = None


@define
class ExperimentSchema:
    name: str
    uuid: str
    created_at: str
    metadata: Optional[Dict[str, str]] = None


@define
class EvaluationResultSchema(EvaluationResult):
    id: Optional[int] = None


@define
class TraceStatsSchema:
    trace_id: str
    latency: Optional[float] = 0.0
    input_tokens: Optional[int] = 0
    output_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0
    cost: Optional[float] = None
    scores: Optional[List[EvaluationResultSchema]] = field(factory=list)


@define
class ExperimentStatsSchema:
    parent_trace_stats: List[TraceStatsSchema] = field(factory=list)

    @property
    def avg_scores(self) -> Dict[str, float]:
        accumulators = {}
        counts = {}
        for trace_stat in self.parent_trace_stats:
            for score in trace_stat.scores:
                accumulators[score.name] = accumulators.get(score.name, 0.0) + score.score
                counts[score.name] = counts.get(score.name, 0) + 1
        return {name: accumulators[name] / counts[name] for name in accumulators}

    def cumulative_avg_score(self) -> float:
        """Returns the average score across all evals."""
        scores = [score.score for trace_stat in self.parent_trace_stats for score in trace_stat.scores]
        return sum(scores) / len(scores) if scores else 0.0

    def avg_score(self, score_name: str) -> float:
        """Returns the average score for a given eval."""
        scores = [score.score for trace_stat in self.parent_trace_stats for score in trace_stat.scores if score.name == score_name]
        return sum(scores) / len(scores) if scores else 0.0


class UpdateTraceScenario(str, Enum):
    RESULT: str = "result"
    ERROR: str = "error"
    CHAIN: str = "chain"
    USAGE: str = "usage"
    OPENAICONFIG: str = "openaiconfig"
    LANGCHAIN_CHILD: str = "langchain_child"


@define
class CreateGetProjectSchema:
    name: str


@define
class ProjectSchema(CreateGetProjectSchema):
    uuid: str
    created_at: str


@define
class CreateGetProjectResponseSchema(ProjectSchema):
    was_created: bool


@define
class TestCase:
    id: int
    test_case_collection_id: int
    inputs: Dict[str, str] = field(factory=dict)
    target: Optional[str] = None
    tags: List[str] = field(factory=list)


class TestCaseResult(list):
    def __init__(self, test_cases: Union[TestCase, List[TestCase]]):
        super().__init__([test_cases] if isinstance(test_cases, TestCase) else test_cases)

    def __getitem__(self, key):
        result = super().__getitem__(key)
        if isinstance(key, slice):
            return TestCaseResult(result)
        return result

    def get_all_test_case_inputs(self) -> List[Dict[str, str]]:
        return [case.inputs for case in self]

    def get_all_test_case_targets(self) -> List[str]:
        return [case.target for case in self if case.target is not None]

    def get_all_test_inputs_and_targets_dict(self) -> List[Dict[str, Any]]:
        return [{"inputs": case.inputs, "target": case.target} for case in self]


@define
class TestCaseCollection:
    id: int
    name: str
    created_at: str
    last_updated_at: str
    column_names: List[str] = field(factory=list)
    test_cases: Dict[int, TestCase] = field(factory=dict)

    @property
    def testcases(self) -> TestCaseResult:
        return TestCaseResult(list(self.test_cases.values()))

    def __getitem__(self, key):
        return self.testcases[key]

    def filter_testcases(self, **kwargs) -> TestCaseResult:
        def matches_criteria(case: TestCase) -> bool:
            for key, value in kwargs.items():
                if key == "inputs":
                    if isinstance(value, dict):
                        if not all(case.inputs.get(k) == v for k, v in value.items()):
                            return False
                    elif isinstance(value, list):
                        for input_filter in value:
                            input_key, condition_func = input_filter
                            if input_key not in case.inputs or not condition_func(case.inputs[input_key]):
                                return False
                elif key == "tags":
                    if isinstance(value, dict):
                        match_type = value.get("match", "any")
                        tags_to_match = value.get("tags", [])
                        if match_type == "all":
                            if not all(tag in case.tags for tag in tags_to_match):
                                return False
                        else:  # 'any'
                            if not any(tag in case.tags for tag in tags_to_match):
                                return False
                    elif isinstance(value, list):
                        if not any(tag in case.tags for tag in value):
                            return False
                elif key == "target":
                    if callable(value):
                        if not value(case.target):
                            return False
                    elif case.target != value:
                        return False
                elif not hasattr(case, key) or getattr(case, key) != value:
                    return False
            return True

        filtered_cases = [case for case in self.test_cases.values() if matches_criteria(case)]
        return TestCaseResult(filtered_cases)

    def get_all_test_case_inputs(self) -> Iterable[Dict[str, str]]:
        return (test_case.inputs for test_case in self.test_cases.values())

    def num_test_cases(self) -> int:
        return len(self.test_cases)

    def get_all_test_case_targets(self) -> Iterable[str]:
        return (test_case.target for test_case in self.test_cases.values())

    def get_all_test_inputs_and_targets_tuple(self) -> Iterable[Tuple[Dict[str, str], Optional[str]]]:
        return ((test_case.inputs, test_case.target) for test_case in self.test_cases.values())

    def get_all_test_inputs_and_targets_dict(self) -> Iterable[Dict[str, str]]:
        return ({**test_case.inputs, "target": test_case.target} for test_case in self.test_cases.values())

    def write_to_finetune_jsonl(self, file_path: str):
        """Converts dataset to finetune jsonl format and writes to file_path."""
        jsonl_rows = []
        for inputs, target in self.get_all_test_inputs_and_targets_tuple():
            messages = json.loads(inputs["messages"])
            try:
                function_call = json.loads(target)
                if isinstance(function_call, List):
                    function_call = function_call[0]
                if "arguments" not in function_call:
                    # tool use format, need to convert
                    function_call = function_call["function"]
                function_call["arguments"] = json.dumps(function_call["arguments"])
                assistant_response = {
                    "role": "assistant",
                    "function_call": function_call,
                }
            except json.JSONDecodeError:
                assistant_response = {"role": "assistant", "content": target}
            messages.append(assistant_response)
            converted_row = {"messages": messages}
            if functions := inputs.get("functions", None):
                if loaded_functions := json.loads(functions):
                    converted_row["functions"] = loaded_functions
            jsonl_rows.append(converted_row)

        data = [json.dumps(line) for line in jsonl_rows]
        with open(file_path, "w", encoding="utf-8") as f:
            f.write("\n".join(data))


@define
class CreateTestCase:
    inputs: Dict[str, str]
    target: Optional[str] = None
    tags: List[str] = field(factory=list)


@define
class CreateTestCases:
    id: Optional[int] = None
    name: Optional[str] = None
    test_cases: List[CreateTestCase] = field(factory=list)

    @validators.optional
    def id_or_name_is_set(self, attribute, value):
        if not (self.id or self.name):
            raise ValueError("One of id or name must be set.")


@define
class CreateTestCaseCollection(CreateTestCases):
    # column names excluding reserved names, target and tags
    column_names: List[str] = field(factory=list)


@define
class UpdateTestCase:
    inputs: Optional[Dict[str, Any]] = None
    target: Optional[Union[int, float, str, bool]] = None
    tags: Optional[List[str]] = None


class ExperimentStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


@define
class FinishExperimentRequestSchema:
    status: ExperimentStatus
    dataset_level_stats: Optional[List[EvaluationResult]] = field(factory=list)


@define
class ListExperimentUUIDsFilters:
    project_name: Optional[str] = None
    metadata_filter: Optional[Dict[str, Any]] = None
    experiment_name_filter: Optional[str] = None
    run_name_filter: Optional[str] = None
    experiment_uuids: Optional[List[str]] = None


class StatisticOperation(str, Enum):
    MEAN = "mean"
    MEDIAN = "median"
    VARIANCE = "variance"
    STANDARD_DEVIATION = "standard_deviation"
    MIN = "min"
    MAX = "max"
    MSE = "mse"
    MAE = "mae"
    CORRELATION = "correlation"
    SPEARMAN_CORRELATION = "spearman_correlation"
    ACCURACY = "accuracy"
    CUSTOM = "custom"


@define
class ExperimentPinnedStatistic:
    var1: str
    operation: StatisticOperation
    value: float
    var2: Optional[str] = None


@define
class ExperimentWithPinnedStatsSchema:
    name: str
    uuid: str
    created_at: str
    run_name: str
    project_uuid: str
    status: ExperimentStatus
    is_public: bool = False
    metadata: Optional[Dict[str, str]] = None
    pinned_stats: List[ExperimentPinnedStatistic] = []
    num_samples: Optional[int] = None


class FilterOperator(str, Enum):
    EQUALS = "equals"
    NOT_EQUALS = "not_equals"
    LIKE = "like"
    GREATER_THAN_OR_EQUAL = "greater_than_or_equal"
    LESS_THAN_OR_EQUAL = "less_than_or_equal"
    GRATER_THAN = "greater_than"
    LESS_THAN = "less_than"
    IS_NULL = "is_null"
    EXISTS = "exists"
    IN = "in"


@define
class TraceLogFilters:
    filter_field: Optional[str] = None
    filter_key: Optional[str] = None
    filter_operator: Optional[FilterOperator] = None
    filter_value: Optional[str] = None


class TimeRange(str, Enum):
    NA = "na"
    LAST_1_HOUR = "1h"
    LAST_3_HOURS = "3h"
    LAST_6_HOURS = "6h"
    LAST_24_HOURS = "24h"
    LAST_7_DAYS = "7d"
    LAST_1_MONTH = "1m"
    LAST_3_MONTHS = "3m"
    LAST_6_MONTHS = "6m"
    LAST_12_MONTHS = "1y"


@define
class QueryParams:
    project_name: str
    filter_field: Optional[str] = None
    filter_key: Optional[str] = None
    filter_operator: Optional[FilterOperator] = None
    filter_value: Optional[str] = None
    page: int = 1
    page_size: int = 20
    time_range: TimeRange = TimeRange.NA
    status: Optional[str] = None


@define
class PaginatedTraceLogsResponse:
    total: int
    page: int
    total_pages: int
    page_size: int
    results: List[TraceLogTree]



from typing import Callable, Dict, Iterable, List, Optional, Union

import asyncio
import inspect
import logging
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from functools import partial
from urllib.parse import quote

from attrs import define, field
from tqdm import tqdm

from parea import Parea
from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.experiment.dvc import save_results_to_dvc_if_init
from parea.helpers import duplicate_dicts, gen_random_name, is_logging_disabled
from parea.schemas import EvaluationResult
from parea.schemas.models import CreateExperimentRequest, ExperimentSchema, ExperimentStatsSchema, ExperimentStatus, FinishExperimentRequestSchema
from parea.utils.trace_utils import thread_ids_running_evals, trace_data
from parea.utils.universal_encoder import json_dumps

STAT_ATTRS = ["latency", "input_tokens", "output_tokens", "total_tokens", "cost"]
logger = logging.getLogger()


def calculate_avg_std_for_experiment(experiment_stats: ExperimentStatsSchema) -> Dict[str, str]:
    accumulators = defaultdict(float)
    counts = defaultdict(int)
    score_accumulators = defaultdict(float)
    score_counts = defaultdict(int)

    for trace_stat in experiment_stats.parent_trace_stats:
        for attr in STAT_ATTRS:
            value = getattr(trace_stat, attr, None)
            if value is not None:
                accumulators[attr] += value
                counts[attr] += 1

        for score in trace_stat.scores:
            score_accumulators[score.name] += score.score
            score_counts[score.name] += 1

    averages = {attr: "N/A" if counts[attr] == 0 else f"{accumulators[attr] / counts[attr]:.{5 if attr == 'cost' else 2}f}" for attr in accumulators}

    score_averages = {f"{name}": "N/A" if score_counts[name] == 0 else f"{score_accumulators[name] / score_counts[name]:.2f}" for name in score_accumulators}

    averages.update(score_averages)

    return averages


def async_wrapper(fn, **kwargs):
    return asyncio.run(fn(**kwargs))


def apply_dataset_eval(dataset_level_evals: List[Callable]) -> List[EvaluationResult]:
    root_traces = []
    for trace in trace_data.get().values():
        if trace.root_trace_id == trace.trace_id:
            root_traces.append(trace)

    results = []
    for dataset_level_eval in dataset_level_evals:
        try:
            result = dataset_level_eval(root_traces)
        except Exception as e:
            logger.error(f"Error occurred calling dataset level eval function '{dataset_level_eval.__name__}': {e}", exc_info=e)
            continue
        if result is None:
            continue

        if isinstance(result, EvaluationResult):
            results.append(result)
        elif isinstance(result, list):
            results.extend(result)
        else:
            results.append(EvaluationResult(name=dataset_level_eval.__name__, score=result))

    return results


async def experiment(
    experiment_name: str,
    run_name: str,
    data: Union[str, int, Iterable[dict]],
    func: Callable,
    p: Parea,
    experiment_uuid: str,
    n_trials: int = 1,
    dataset_level_evals: Optional[List[Callable]] = None,
    n_workers: int = 10,
    stop_on_error: bool = True,
) -> ExperimentStatsSchema:
    """Creates an experiment and runs the function on the data iterator.
    param experiment_name: The name of the experiment. Used to organize experiments within a project.
    param run_name: The run name of the experiment. This name must be unique across experiment runs.
    param data: The data to run the experiment on. This can be a list of dictionaries,
        a string representing the name of a dataset on Parea or an int representing the id of a dataset on Parea.
        If it is a list of dictionaries, the key "target" is reserved for the target/expected output of that sample.
    param func: The function to run. This function should accept inputs that match the keys of the data field.
    param p: The Parea instance to use for running the experiment.
    param experiment_uuid: The UUID of the experiment. This is used to associate traces with the experiment.
    param n_trials: The number of times to run the experiment on the same data.
    param dataset_level_evals: A list of functions to run on the dataset level. Each function should accept a list of EvaluatedLogs and return a float or a
        EvaluationResult. If a float is returned, the name of the function will be used as the name of the evaluation.
    param n_workers: The number of workers to use for running the experiment.
    param stop_on_error: If True, the experiment will stop running if an exception is raised.
    """
    if isinstance(data, (str, int)):
        print(f"Fetching test collection: {data}")
        test_collection = await p.aget_collection(data)
        len_test_cases = test_collection.num_test_cases()
        print(f"Fetched {test_collection.num_test_cases()} test cases from collection: {data} \n")
        data: Iterable[dict] = test_collection.get_all_test_inputs_and_targets_dict()
    else:
        len_test_cases = len(data) if isinstance(data, list) else 0

    if n_trials > 1:
        data = duplicate_dicts(data, n_trials)
        len_test_cases = len(data) if isinstance(data, list) else 0
        print(f"Running {n_trials} trials of the experiment \n")

    os.environ[PAREA_OS_ENV_EXPERIMENT_UUID] = experiment_uuid

    sem = asyncio.Semaphore(n_workers)

    async def limit_concurrency(sample):
        async with sem:
            sample_copy = deepcopy(sample)
            target = sample_copy.pop("target", None)
            return await func(_parea_target_field=target, **sample_copy)

    def limit_concurrency_sync(sample):
        sample_copy = deepcopy(sample)
        target = sample_copy.pop("target", None)
        return func(_parea_target_field=target, **sample_copy)

    if inspect.iscoroutinefunction(func):
        tasks = [asyncio.ensure_future(limit_concurrency(sample)) for sample in data]
    else:
        executor = ThreadPoolExecutor(max_workers=n_workers)
        loop = asyncio.get_event_loop()
        tasks = [asyncio.ensure_future(loop.run_in_executor(executor, partial(limit_concurrency_sync, sample))) for sample in data]

    status = ExperimentStatus.COMPLETED
    with tqdm(total=len(tasks), desc="Running samples", unit="sample") as pbar:
        try:
            for coro in asyncio.as_completed(tasks):
                try:
                    await coro
                    pbar.update(1)
                except Exception as e:
                    status = ExperimentStatus.FAILED
                    if stop_on_error:
                        print(f"\nExperiment stopped due to an error (note you can deactivate this behavior by setting stop_on_error=False): {str(e)}\n")
                        for task in tasks:
                            task.cancel()
                    else:
                        pbar.update(1)
        except asyncio.CancelledError:
            pass

    await asyncio.sleep(0.2)
    total_evals = len(thread_ids_running_evals.get())
    with tqdm(total=total_evals, dynamic_ncols=True) as pbar:
        while thread_ids_running_evals.get():
            pbar.set_description("Waiting for evaluations to finish")
            pbar.update(total_evals - len(thread_ids_running_evals.get()))
            total_evals = len(thread_ids_running_evals.get())
            await asyncio.sleep(0.5)
        await asyncio.sleep(4)
        pbar.update(total_evals)

    if dataset_level_evals:
        dataset_level_eval_results = apply_dataset_eval(dataset_level_evals)
    else:
        dataset_level_eval_results = []

    experiment_stats: ExperimentStatsSchema = p.finish_experiment(experiment_uuid, FinishExperimentRequestSchema(status=status, dataset_level_stats=dataset_level_eval_results))
    stat_name_to_avg_std = calculate_avg_std_for_experiment(experiment_stats)
    if dataset_level_eval_results:
        stat_name_to_avg_std.update(**{eval_result.name: eval_result.score for eval_result in dataset_level_eval_results})
    print(f"Experiment {experiment_name} Run {run_name} stats:\n{json_dumps(stat_name_to_avg_std, indent=2)}\n\n")
    print(f"View experiment & traces at: https://app.parea.ai/experiments/{quote(experiment_name, safe='')}/{experiment_uuid}\n")
    save_results_to_dvc_if_init(run_name, stat_name_to_avg_std)

    if os.environ.get(PAREA_OS_ENV_EXPERIMENT_UUID, None):
        del os.environ[PAREA_OS_ENV_EXPERIMENT_UUID]

    return experiment_stats


_experiments = []


def data_converter(data: Union[str, int, Iterable[dict]]) -> Union[str, int, Iterable[dict]]:
    if isinstance(data, (str, int)):
        return data
    else:
        for sample in data:
            if isinstance(sample, dict) and "target" in sample and isinstance(sample["target"], dict):
                sample["target"] = json_dumps(sample["target"])
        return data


@define
class Experiment:
    # If your dataset is defined locally it should be an iterable of k/v
    # pairs matching the expected inputs of your function. To reference a dataset you
    # have saved on Parea, use the dataset name as a string or the id as an int.
    data: Union[str, int, Iterable[dict]] = field(converter=data_converter)
    # The function to run. This function should accept inputs that match the keys of the data field.
    func: Callable = field()
    experiment_stats: ExperimentStatsSchema = field(init=False, default=None)
    metadata: Optional[Dict[str, str]] = field(default=None)
    dataset_level_evals: Optional[List[Callable]] = field(default=None)
    p: Parea = field(default=None)
    experiment_name: str = field
    run_name: str = field(init=False)
    experiment_uuid: str = field(init=False, default=None)
    n_workers: int = field(default=10)
    # The number of times to run the experiment on the same data.
    n_trials: int = field(default=1)
    stop_on_error: bool = field(default=True)

    def __attrs_post_init__(self):
        global _experiments
        _experiments.append(self)
        if isinstance(self.data, str):
            if self.metadata is None:
                self.metadata = {"Dataset": self.data}
            else:
                if "Dataset" in self.metadata:
                    raise ValueError("Metadata should not contain a key 'Dataset' when using uploaded dataset (data is a string).")
                self.metadata["Dataset"] = self.data

    def _gen_run_name_if_none(self, name: Optional[str]):
        if not name:
            self.run_name = gen_random_name()
            print(f"Run name set to: {self.run_name}, since a name was not provided.")
        else:
            self.run_name = name

    def run(self, run_name: Optional[str] = None) -> None:
        """Run the experiment and save the results to DVC.
        param run_name: The run name of the experiment. This name must be unique across experiment runs.
        If no run name is provided a memorable name will be generated automatically.
        """
        if is_logging_disabled():
            print("Parea logging is turned off. Experiment can't be run without logging. Set env var TURN_OFF_PAREA_LOGGING to False to enable.")
            return

        try:
            self._gen_run_name_if_none(run_name)
            experiment_schema: ExperimentSchema = self.p.create_experiment(CreateExperimentRequest(name=self.experiment_name, run_name=self.run_name, metadata=self.metadata))
            self.experiment_uuid = experiment_schema.uuid
            self.experiment_stats = asyncio.run(
                experiment(
                    self.experiment_name,
                    self.run_name,
                    self.data,
                    self.func,
                    self.p,
                    self.experiment_uuid,
                    self.n_trials,
                    self.dataset_level_evals,
                    self.n_workers,
                    self.stop_on_error,
                )
            )
        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error running experiment: {e}")

    async def arun(self, run_name: Optional[str] = None) -> None:
        """Run the experiment and save the results to DVC.
        param run_name: The run name of the experiment. This name must be unique across experiment runs.
        If no run name is provided a memorable name will be generated automatically.
        """
        if is_logging_disabled():
            print("Parea logging is turned off. Experiment can't be run without logging. Set env var TURN_OFF_PAREA_LOGGING to False to enable.")
            return

        try:
            self._gen_run_name_if_none(run_name)
            experiment_schema: ExperimentSchema = await self.p.acreate_experiment(
                CreateExperimentRequest(name=self.experiment_name, run_name=self.run_name, metadata=self.metadata)
            )
            self.experiment_uuid = experiment_schema.uuid
            self.experiment_stats = await experiment(
                self.experiment_name, self.run_name, self.data, self.func, self.p, self.experiment_uuid, self.n_trials, self.dataset_level_evals, self.n_workers, self.stop_on_error
            )
        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error running experiment: {e}")

    @property
    def avg_scores(self) -> Dict[str, float]:
        """Returns the average score across all evals."""
        return self.experiment_stats.avg_scores if self.experiment_stats else {}



from typing import Callable, Optional, Union

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def answer_matches_target_llm_grader_factory(
    question_field: Optional[str] = "question",
    model: Optional[str] = "gpt-4",
    is_azure: Optional[bool] = False,
) -> Callable[[Log], Union[float, None]]:
    """Quantifies how much the generated answer matches the ground truth / target."""

    def answer_matches_target_llm_grader(log: Log) -> Union[float, None]:
        question = log.inputs[question_field]
        output = log.output
        if (target := log.target) is None:
            return None
        response = call_openai(
            model=model,
            messages=[
                {"role": "system", "content": "You are CompareGPT, a machine to verify the groundedness of predictions. Answer with " "only yes/no."},
                {
                    "role": "user",
                    "content": f"""You are given a question, the corresponding ground-truth answer and a prediction from a model. Compare the "Ground-truth answer" and the "Prediction" to determine whether the prediction correctly answers the question. All information in the ground-truth answer must be present in the prediction, including numbers and dates. You must answer "no" if there are any specific details in the ground-truth answer that are not mentioned in the prediction. There should be no contradicting statements in the prediction. The prediction may contain extra information. If the prediction states something as a possibility, treat it as a definitive answer.

Question: {question}
Ground-truth answer: {target}
Prediction: {output}

CompareGPT response:""",
                },
            ],
            temperature=0.0,
            is_azure=is_azure,
        )
        return float("yes" in response.lower())

    return answer_matches_target_llm_grader



from typing import Any, Dict, List, Optional, Tuple, Union

import functools

import cohere
from attrs import asdict, define
from cohere import ApiMetaBilledUnits, NonStreamedChatResponse, RerankResponse

from parea.constants import COHERE_MODEL_INFO, COHERE_SEARCH_MODELS
from parea.schemas import Message, Role
from parea.utils.universal_encoder import json_dumps

DEFAULT_MODEL = "command-r-plus"
DEFAULT_TEMPERATURE = 0.3
DEFAULT_P = 0.75


@define
class CohereOutput:
    text: Optional[str] = None
    citations: Optional[str] = None
    documents: Optional[str] = None
    search_queries: Optional[str] = None
    search_results: Optional[str] = None


def chat_history_to_messages(result: NonStreamedChatResponse, **kwargs) -> list[Message]:
    messages: list[Message] = []
    if sys_message := kwargs.get("preamble", ""):
        messages.append(Message(content=sys_message, role=Role.system))
    if history := kwargs.get("chat_history", []):
        messages.extend(to_messages(history))

    messages.extend(to_messages([m.dict() for m in result.chat_history]))
    return messages


def to_messages(chat_history: List[Union[Dict, cohere.Message]]) -> List[Message]:
    role_map = {"USER": Role.user, "CHATBOT": Role.assistant, "SYSTEM": Role.system, "TOOL": Role.tool}

    def process_message(message: Union[Dict, cohere.Message]) -> Message:
        if isinstance(message, dict):
            role = role_map.get(message["role"], Role.user)
            content = message.get("message", "")
            tool_calls = message.get("tool_calls") or message.get("tool_results")
        else:  # cohere.Message
            role = role_map.get(message.role, Role.user)
            content = "" if role == Role.tool else message.message
            tool_calls = getattr(message, "tool_calls", None) or getattr(message, "tool_results", None)

        if tool_calls:
            tc = json_dumps([t.dict() if hasattr(t, "dict") else t for t in tool_calls])
            content = tc if role == Role.tool or not content else json_dumps({"message": content, "tool_calls": tc})

        return Message(content=content, role=role)

    return list(map(process_message, chat_history))


@functools.lru_cache(maxsize=128)
def compute_cost(prompt_tokens: int, completion_tokens: int, search_units: int, is_search_model: bool, model: str) -> float:
    cost_per_token = COHERE_MODEL_INFO.get(model, {"prompt": 0, "completion": 0})
    cost = ((prompt_tokens * cost_per_token["prompt"]) + (completion_tokens * cost_per_token["completion"])) / 1_000_000
    if is_search_model:
        cost += search_units * cost_per_token.get("search", 0) / 1_000
    cost = round(cost, 10)
    return cost


def get_usage_stats(result: Optional[NonStreamedChatResponse | RerankResponse], model: str) -> Tuple[int, int, float]:
    bu: Optional[ApiMetaBilledUnits] = result.meta.billed_units if result else None
    if not bu:
        return 0, 0, 0.0
    prompt_tokens = bu.input_tokens or 0
    completion_tokens = bu.output_tokens or 0
    search_units = bu.search_units or 0
    is_search_model: bool = model in COHERE_SEARCH_MODELS
    cost = compute_cost(prompt_tokens, completion_tokens, search_units, is_search_model, model)
    return prompt_tokens, completion_tokens, cost


def get_output(result: Optional[NonStreamedChatResponse | RerankResponse]) -> str:
    if not result:
        return ""

    if isinstance(result, RerankResponse):
        output = CohereOutput(documents=cohere_json_list(result.results) if result.results else None)
        return json_dumps(asdict(output))

    text = result.text or cohere_json_list(result.tool_calls)
    output = CohereOutput(
        text=text,
        citations=cohere_json_list(result.citations) if result.citations else None,
        documents=cohere_json_list(result.documents) if result.documents else None,
        search_queries=cohere_json_list(result.search_queries) if result.search_queries else None,
        search_results=cohere_json_list(result.search_results) if result.search_results else None,
    )
    return json_dumps(asdict(output))


def cohere_json_list(obj: Any) -> str:
    out = []
    for o in obj or []:
        if isinstance(o, dict):
            out.append(o)
        else:
            out.append(o.dict())
    return json_dumps(out)



from typing import List, Union

from collections import defaultdict

from parea.schemas import EvaluatedLog, EvaluationResult


def balanced_acc_factory(score_name: str):
    def balanced_acc(logs: List[EvaluatedLog]) -> Union[EvaluationResult, None]:
        correct = defaultdict(int)
        total = defaultdict(int)
        for log in logs:
            if (eval_result := log.get_score(score_name)) is not None:
                correct[log.target] += int(eval_result.score)
                total[log.target] += 1
        recalls = [correct[key] / total[key] for key in correct]

        if len(recalls) == 0:
            return None

        return EvaluationResult(name=f"balanced_acc_{score_name}", score=sum(recalls) / len(recalls))

    return balanced_acc



from typing import Dict, Union

import os

from dotenv import load_dotenv

load_dotenv()


def str2bool(v):
    if isinstance(v, bool):
        return v
    if not v or not isinstance(v, str):
        return False
    return v.lower() in ("yes", "true", "t", "1")


TURN_OFF_PAREA_LOGGING = str2bool(os.getenv("TURN_OFF_PAREA_LOGGING", False))
TURN_OFF_PAREA_EVAL_LOGGING = str2bool(os.getenv("TURN_OFF_PAREA_EVAL_LOGGING", False))

PAREA_OS_ENV_EXPERIMENT_UUID = "_PAREA_EXPERIMENT_UUID"
PAREA_DVC_DIR = ".parea"
PAREA_DVC_METRICS_FILE = str(os.path.join(PAREA_DVC_DIR, "metrics.json"))
PAREA_DVC_YAML_FILE = str(os.path.join(PAREA_DVC_DIR, "dvc.yaml"))

CHUNK_DONE_SENTINEL = "data: [DONE]"

OPENAI_MODEL_INFO: Dict[str, Dict[str, Union[float, int, Dict[str, int]]]] = {
    "gpt-3.5-turbo-0301": {
        "prompt": 1.5,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-3.5-turbo-0613": {
        "prompt": 1.5,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-3.5-turbo-16k": {
        "prompt": 3.0,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 16385, "max_prompt_tokens": 16385},
    },
    "gpt-3.5-turbo-16k-0301": {
        "prompt": 3.0,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 16385, "max_prompt_tokens": 16385},
    },
    "gpt-3.5-turbo-16k-0613": {
        "prompt": 3.0,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 16385, "max_prompt_tokens": 16385},
    },
    "gpt-3.5-turbo-1106": {
        "prompt": 1.0,
        "completion": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-3.5-turbo-0125": {
        "prompt": 0.5,
        "completion": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 16385},
    },
    "gpt-3.5-turbo": {
        "prompt": 0.5,
        "completion": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 16385},
    },
    "gpt-3.5-turbo-instruct": {
        "prompt": 1.5,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-4": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "gpt-4-0314": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "gpt-4-0613": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "gpt-4-32k": {
        "prompt": 60.0,
        "completion": 120.0,
        "token_limit": {"max_completion_tokens": 32768, "max_prompt_tokens": 32768},
    },
    "gpt-4-32k-0314": {
        "prompt": 60.0,
        "completion": 120.0,
        "token_limit": {"max_completion_tokens": 32768, "max_prompt_tokens": 32768},
    },
    "gpt-4-32k-0613": {
        "prompt": 60.0,
        "completion": 120.0,
        "token_limit": {"max_completion_tokens": 32768, "max_prompt_tokens": 32768},
    },
    "gpt-4-vision-preview": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-1106-vision-preview": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-turbo-preview": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-1106-preview": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-0125-preview": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-turbo": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-turbo-2024-04-09": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o-2024-05-13": {
        "prompt": 5.0,
        "completion": 15.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o": {
        "prompt": 5.0,
        "completion": 15.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o-2024-08-06": {
        "prompt": 2.5,
        "completion": 10.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o-mini": {
        "prompt": 0.15,
        "completion": 0.6,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o-mini-2024-07-18": {
        "prompt": 0.15,
        "completion": 0.6,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
}
AZURE_MODEL_INFO: Dict[str, Dict[str, Union[float, int, Dict[str, int]]]] = {
    "gpt-35-turbo": {
        "prompt": 1.5,
        "completion": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-35-turbo-0301": {
        "prompt": 1.5,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-35-turbo-0613": {
        "prompt": 1.5,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "gpt-35-turbo-16k": {
        "prompt": 3.0,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 16384, "max_prompt_tokens": 16384},
    },
    "gpt-35-turbo-16k-0301": {
        "prompt": 3.0,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 16384, "max_prompt_tokens": 16384},
    },
    "gpt-35-turbo-16k-0613": {
        "prompt": 3.0,
        "completion": 4.0,
        "token_limit": {"max_completion_tokens": 16384, "max_prompt_tokens": 16384},
    },
    "gpt-4": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "gpt-4-0314": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "gpt-4-0613": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "gpt-4-32k": {
        "prompt": 60.0,
        "completion": 120.0,
        "token_limit": {"max_completion_tokens": 32768, "max_prompt_tokens": 32768},
    },
    "gpt-4-32k-0314": {
        "prompt": 60.0,
        "completion": 120.0,
        "token_limit": {"max_completion_tokens": 32768, "max_prompt_tokens": 32768},
    },
    "gpt-4-32k-0613": {
        "prompt": 60.0,
        "completion": 120.0,
        "token_limit": {"max_completion_tokens": 32768, "max_prompt_tokens": 32768},
    },
    "gpt-4-1106-preview": {
        "prompt": 10.0,
        "completion": 20.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-vision": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-1106-vision-preview": {
        "prompt": 30.0,
        "completion": 60.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-35-turbo-instruct": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-turbo": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4-turbo-2024-04-09": {
        "prompt": 10.0,
        "completion": 30.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o-2024-05-13": {
        "prompt": 5.0,
        "completion": 15.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "gpt-4o": {
        "prompt": 5.0,
        "completion": 15.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
}

instant_pricing = {"prompt": 1.63, "completion": 5.51}
claude_pricing = {"prompt": 8.00, "completion": 24.00}
ANTHROPIC_MODEL_INFO: Dict[str, Dict[str, Union[float, int, Dict[str, float]]]] = {
    "claude-instant-1.1": {
        "token_limit": {
            "max_prompt_tokens": 9000,
            "max_completion_tokens": 4096,
        },
        **instant_pricing,
    },
    "claude-instant-1": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **instant_pricing,
    },
    "claude-instant-1.2": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **instant_pricing,
    },
    "claude-instant-1-100k": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **instant_pricing,
    },
    "claude-instant-1.1-100k": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **instant_pricing,
    },
    "claude-1": {
        "token_limit": {
            "max_prompt_tokens": 9000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-2": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-1-100k": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-1.2": {
        "token_limit": {
            "max_prompt_tokens": 9000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-1.3": {
        "token_limit": {
            "max_prompt_tokens": 9000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-1.3-100k": {
        "token_limit": {
            "max_prompt_tokens": 100000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-2.1": {
        "token_limit": {
            "max_prompt_tokens": 200000,
            "max_completion_tokens": 4096,
        },
        **claude_pricing,
    },
    "claude-3-opus-20240229": {
        "token_limit": {
            "max_prompt_tokens": 200000,
            "max_completion_tokens": 4096,
        },
        "prompt": 15.00,
        "completion": 75.00,
    },
    "claude-3-sonnet-20240229": {
        "token_limit": {"max_prompt_tokens": 200000, "max_completion_tokens": 4096},
        "prompt": 3.00,
        "completion": 15.00,
    },
    "claude-3-haiku-20240307": {
        "token_limit": {"max_prompt_tokens": 200000, "max_completion_tokens": 4096},
        "prompt": 0.25,
        "completion": 1.25,
    },
    "claude-3-5-sonnet-20240620": {
        "token_limit": {"max_prompt_tokens": 200000, "max_completion_tokens": 4096},
        "prompt": 3.00,
        "completion": 15.00,
    },
}
COHERE_MODEL_INFO: Dict[str, Dict[str, Union[float, int, Dict[str, int]]]] = {
    "command-r-plus": {
        "prompt": 3.0,
        "completion": 15.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "command-r": {
        "prompt": 0.5,
        "completion": 1.5,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 128000},
    },
    "command": {
        "prompt": 1.0,
        "completion": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "command-nightly": {
        "prompt": 1.0,
        "completion": 2.0,
        "token_limit": {"max_completion_tokens": 128000, "max_prompt_tokens": 128000},
    },
    "command-light": {
        "prompt": 0.3,
        "completion": 0.6,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "command-light-nightly": {
        "prompt": 0.3,
        "completion": 0.6,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "c4ai-aya-23": {
        "prompt": 0.0,
        "completion": 0.0,
        "token_limit": {"max_completion_tokens": 8192, "max_prompt_tokens": 8192},
    },
    "rerank-english-v3.0": {
        "prompt": 0,
        "completion": 0,
        # $ per 1K
        "search": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "rerank-multilingual-v3.0": {
        "prompt": 0,
        "completion": 0,
        # $ per 1K
        "search": 2.0,
        "token_limit": {"max_completion_tokens": 4096, "max_prompt_tokens": 4096},
    },
    "rerank-english-v2.0": {
        "prompt": 0,
        "completion": 0,
        # $ per 1K
        "search": 1.0,
        "token_limit": {"max_completion_tokens": 512, "max_prompt_tokens": 512},
    },
    "rerank-multilingual-v2.0": {
        "prompt": 0,
        "completion": 0,
        # $ per 1K
        "search": 1.0,
        "token_limit": {"max_completion_tokens": 512, "max_prompt_tokens": 512},
    },
}
COHERE_SEARCH_MODELS = {
    "rerank-english-v3.0",
    "rerank-multilingual-v3.0",
    "rerank-english-v2.0",
    "rerank-multilingual-v2.0",
}
ALL_NON_AZURE_MODELS_INFO = {**OPENAI_MODEL_INFO, **ANTHROPIC_MODEL_INFO, **COHERE_MODEL_INFO}

NOUNS = (
    "abac",
    "abbs",
    "aces",
    "acid",
    "acne",
    "acre",
    "acts",
    "ados",
    "adze",
    "afro",
    "agas",
    "aged",
    "ages",
    "agio",
    "agma",
    "airs",
    "airt",
    "aits",
    "akes",
    "alap",
    "albs",
    "alga",
    "ally",
    "alto",
    "amah",
    "ambo",
    "amie",
    "amyl",
    "ankh",
    "apex",
    "aqua",
    "arcs",
    "areg",
    "aria",
    "aril",
    "arks",
    "army",
    "auks",
    "aune",
    "aura",
    "awls",
    "awns",
    "axon",
    "azan",
    "baby",
    "bade",
    "bael",
    "bags",
    "bait",
    "ball",
    "banc",
    "bang",
    "bani",
    "barb",
    "bark",
    "bate",
    "bats",
    "bawl",
    "beak",
    "bean",
    "beep",
    "belt",
    "berk",
    "beth",
    "bias",
    "bice",
    "bids",
    "bind",
    "bise",
    "bish",
    "bite",
    "boar",
    "boat",
    "body",
    "boff",
    "bold",
    "boll",
    "bolo",
    "bomb",
    "bond",
    "book",
    "boor",
    "boot",
    "bort",
    "bosk",
    "bots",
    "bott",
    "bout",
    "bras",
    "bree",
    "brig",
    "brio",
    "buck",
    "buhl",
    "bump",
    "bunk",
    "bunt",
    "buoy",
    "byes",
    "byte",
    "cane",
    "cant",
    "caps",
    "care",
    "cart",
    "cats",
    "cedi",
    "ceps",
    "cere",
    "chad",
    "cham",
    "chat",
    "chay",
    "chic",
    "chin",
    "chis",
    "chiv",
    "choc",
    "chow",
    "chum",
    "ciao",
    "cigs",
    "clay",
    "clip",
    "clog",
    "coal",
    "coat",
    "code",
    "coed",
    "cogs",
    "coho",
    "cole",
    "cols",
    "colt",
    "conk",
    "cons",
    "cony",
    "coof",
    "cook",
    "cool",
    "coos",
    "corm",
    "cors",
    "coth",
    "cows",
    "coze",
    "crag",
    "craw",
    "cree",
    "crib",
    "cuds",
    "cull",
    "cult",
    "curb",
    "curn",
    "curs",
    "cusp",
    "cuss",
    "cwms",
    "cyma",
    "cyst",
    "dabs",
    "dado",
    "daff",
    "dais",
    "daks",
    "damn",
    "dams",
    "darg",
    "dart",
    "data",
    "dawk",
    "dawn",
    "daws",
    "daze",
    "dean",
    "debs",
    "debt",
    "deep",
    "dees",
    "dele",
    "delf",
    "dent",
    "deys",
    "dhow",
    "digs",
    "dirk",
    "dita",
    "diva",
    "divs",
    "doek",
    "doge",
    "dogs",
    "dogy",
    "dohs",
    "doit",
    "dole",
    "doll",
    "dolt",
    "dona",
    "dook",
    "door",
    "dops",
    "doss",
    "doxy",
    "drab",
    "drop",
    "drum",
    "duad",
    "duct",
    "duff",
    "duke",
    "dunk",
    "dunt",
    "ears",
    "ease",
    "eggs",
    "eild",
    "emeu",
    "emus",
    "envy",
    "epha",
    "eric",
    "erns",
    "esne",
    "esse",
    "ewes",
    "expo",
    "eyas",
    "eyot",
    "eyry",
    "fare",
    "farl",
    "farm",
    "feds",
    "feel",
    "fees",
    "feme",
    "fess",
    "fibs",
    "fids",
    "fils",
    "firm",
    "fish",
    "flab",
    "flap",
    "flea",
    "flew",
    "flex",
    "flip",
    "flit",
    "flus",
    "flux",
    "foil",
    "fond",
    "food",
    "fool",
    "ford",
    "fore",
    "frit",
    "friz",
    "froe",
    "funs",
    "furl",
    "fuss",
    "fuzz",
    "gaby",
    "gaff",
    "gale",
    "gang",
    "gaol",
    "gape",
    "gash",
    "gaur",
    "gaze",
    "gear",
    "genu",
    "gest",
    "geum",
    "ghat",
    "gigs",
    "gimp",
    "gird",
    "girl",
    "glee",
    "glen",
    "glia",
    "glop",
    "gnat",
    "goad",
    "goaf",
    "gobs",
    "gonk",
    "good",
    "goos",
    "gore",
    "gram",
    "gray",
    "grig",
    "grip",
    "grot",
    "grub",
    "gude",
    "gula",
    "gulf",
    "guns",
    "gust",
    "gyms",
    "gyro",
    "hack",
    "haet",
    "hajj",
    "hake",
    "half",
    "halm",
    "hard",
    "harl",
    "hask",
    "hate",
    "he'd",
    "heck",
    "heel",
    "heir",
    "help",
    "hems",
    "here",
    "hill",
    "hips",
    "hits",
    "hobo",
    "hock",
    "hogs",
    "hold",
    "holy",
    "hood",
    "hoot",
    "hope",
    "horn",
    "hose",
    "hour",
    "hows",
    "huck",
    "hugs",
    "huia",
    "hulk",
    "hull",
    "hunk",
    "hunt",
    "huts",
    "hymn",
    "ibex",
    "ices",
    "iglu",
    "impi",
    "inks",
    "inti",
    "ions",
    "iota",
    "iron",
    "jabs",
    "jags",
    "jake",
    "jass",
    "jato",
    "jaws",
    "jean",
    "jeer",
    "jerk",
    "jest",
    "jiao",
    "jigs",
    "jill",
    "jinn",
    "jird",
    "jive",
    "jock",
    "joey",
    "jogs",
    "joss",
    "jota",
    "jots",
    "juba",
    "jube",
    "judo",
    "jump",
    "junk",
    "jura",
    "juts",
    "jynx",
    "kago",
    "kail",
    "kaka",
    "kale",
    "kana",
    "keek",
    "keep",
    "kefs",
    "kegs",
    "kerf",
    "kern",
    "keys",
    "kibe",
    "kick",
    "kids",
    "kifs",
    "kill",
    "kina",
    "kind",
    "kine",
    "kite",
    "kiwi",
    "knap",
    "knit",
    "koas",
    "kobs",
    "kyat",
    "lack",
    "lahs",
    "lair",
    "lama",
    "lamb",
    "lame",
    "lats",
    "lava",
    "lays",
    "leaf",
    "leak",
    "leas",
    "lees",
    "leks",
    "leno",
    "libs",
    "lich",
    "lick",
    "lien",
    "lier",
    "lieu",
    "life",
    "lift",
    "limb",
    "line",
    "link",
    "linn",
    "lira",
    "loft",
    "loge",
    "loir",
    "long",
    "loof",
    "look",
    "loot",
    "lore",
    "loss",
    "lots",
    "loup",
    "love",
    "luce",
    "ludo",
    "luke",
    "lulu",
    "lure",
    "lush",
    "magi",
    "maid",
    "main",
    "mako",
    "male",
    "mana",
    "many",
    "mart",
    "mash",
    "mast",
    "mate",
    "math",
    "mats",
    "matt",
    "maul",
    "maya",
    "mays",
    "meal",
    "mean",
    "meed",
    "mela",
    "mene",
    "mere",
    "merk",
    "mesh",
    "mete",
    "mice",
    "milo",
    "mime",
    "mina",
    "mine",
    "mirk",
    "miss",
    "mobs",
    "moit",
    "mold",
    "molt",
    "mome",
    "moms",
    "monk",
    "moot",
    "mope",
    "more",
    "morn",
    "mows",
    "moxa",
    "much",
    "mung",
    "mush",
    "muss",
    "myth",
    "name",
    "nard",
    "nark",
    "nave",
    "navy",
    "neck",
    "newt",
    "nibs",
    "nims",
    "nine",
    "nock",
    "noil",
    "noma",
    "nosh",
    "nowt",
    "nuke",
    "oafs",
    "oast",
    "oats",
    "obit",
    "odor",
    "okra",
    "omer",
    "oner",
    "ones",
    "orcs",
    "ords",
    "orfe",
    "orle",
    "ossa",
    "outs",
    "over",
    "owls",
    "pail",
    "pall",
    "palp",
    "pams",
    "pang",
    "pans",
    "pant",
    "paps",
    "pate",
    "pats",
    "paws",
    "pear",
    "peba",
    "pech",
    "pecs",
    "peel",
    "peer",
    "pees",
    "pein",
    "peri",
    "phon",
    "pice",
    "pita",
    "pith",
    "play",
    "plop",
    "plot",
    "plow",
    "plug",
    "plum",
    "polo",
    "pomp",
    "pond",
    "pons",
    "pony",
    "poof",
    "pope",
    "poss",
    "pots",
    "pour",
    "prad",
    "prat",
    "prep",
    "prob",
    "prof",
    "prow",
    "puck",
    "puds",
    "puke",
    "puku",
    "pump",
    "puns",
    "pupa",
    "purl",
    "pyre",
    "quad",
    "quay",
    "quey",
    "quiz",
    "raid",
    "rail",
    "rain",
    "raja",
    "rale",
    "rams",
    "rand",
    "rant",
    "raps",
    "rasp",
    "razz",
    "rede",
    "reef",
    "reif",
    "rein",
    "repp",
    "rial",
    "ribs",
    "rick",
    "rift",
    "rill",
    "rime",
    "rims",
    "ring",
    "rins",
    "rise",
    "rite",
    "rits",
    "roam",
    "robe",
    "rods",
    "roma",
    "rook",
    "rort",
    "rotl",
    "roup",
    "roux",
    "rube",
    "rubs",
    "ruby",
    "rues",
    "rugs",
    "ruin",
    "runs",
    "ryas",
    "sack",
    "sacs",
    "saga",
    "sail",
    "sale",
    "salp",
    "salt",
    "sand",
    "sang",
    "sash",
    "saut",
    "says",
    "scab",
    "scow",
    "scud",
    "scup",
    "scut",
    "seal",
    "seam",
    "sech",
    "seed",
    "seep",
    "seer",
    "self",
    "sena",
    "send",
    "sera",
    "sere",
    "shad",
    "shah",
    "sham",
    "shay",
    "shes",
    "ship",
    "shoe",
    "sick",
    "sida",
    "sign",
    "sike",
    "sima",
    "sine",
    "sing",
    "sinh",
    "sink",
    "sins",
    "site",
    "size",
    "skat",
    "skin",
    "skip",
    "skis",
    "slaw",
    "sled",
    "slew",
    "sley",
    "slob",
    "slue",
    "slug",
    "smut",
    "snap",
    "snib",
    "snip",
    "snob",
    "snog",
    "snot",
    "snow",
    "snub",
    "snug",
    "soft",
    "soja",
    "soke",
    "song",
    "sons",
    "sook",
    "sorb",
    "sori",
    "souk",
    "soul",
    "sous",
    "soya",
    "spit",
    "stay",
    "stew",
    "stir",
    "stob",
    "stud",
    "suds",
    "suer",
    "suit",
    "sumo",
    "sums",
    "sups",
    "suqs",
    "suss",
    "sway",
    "syce",
    "synd",
    "taal",
    "tach",
    "taco",
    "tads",
    "taka",
    "tale",
    "tamp",
    "tams",
    "tang",
    "tans",
    "tape",
    "tare",
    "taro",
    "tarp",
    "tart",
    "tass",
    "taus",
    "teat",
    "teds",
    "teff",
    "tegu",
    "tell",
    "term",
    "thar",
    "thaw",
    "tics",
    "tier",
    "tiff",
    "tils",
    "tilt",
    "tint",
    "tipi",
    "tire",
    "tirl",
    "toby",
    "tods",
    "toea",
    "toff",
    "toga",
    "toil",
    "toke",
    "tola",
    "tole",
    "tomb",
    "toms",
    "torc",
    "tors",
    "tort",
    "tosh",
    "tote",
    "tret",
    "trey",
    "trio",
    "trug",
    "tuck",
    "tugs",
    "tule",
    "tune",
    "tuns",
    "tuts",
    "tyke",
    "tyne",
    "typo",
    "ulna",
    "umbo",
    "unau",
    "unit",
    "upas",
    "user",
    "uvea",
    "vacs",
    "vane",
    "vang",
    "vans",
    "vara",
    "vase",
    "veep",
    "veer",
    "vega",
    "veil",
    "vela",
    "vent",
    "vies",
    "view",
    "vina",
    "vine",
    "vise",
    "vlei",
    "volt",
    "vows",
    "wads",
    "waft",
    "wage",
    "wain",
    "walk",
    "want",
    "wart",
    "wave",
    "waws",
    "weal",
    "wean",
    "weds",
    "weep",
    "weft",
    "weir",
    "weka",
    "weld",
    "wens",
    "weys",
    "whap",
    "whey",
    "whin",
    "whit",
    "whop",
    "wide",
    "wife",
    "wind",
    "wine",
    "wino",
    "wins",
    "wire",
    "wise",
    "woes",
    "wont",
    "wool",
    "work",
    "worm",
    "wort",
    "yack",
    "yank",
    "yapp",
    "yard",
    "yate",
    "yawl",
    "yegg",
    "yell",
    "yeuk",
    "yews",
    "yips",
    "yobs",
    "yogi",
    "yoke",
    "yolk",
    "yoni",
    "zack",
    "zags",
    "zest",
    "zhos",
    "zigs",
    "zila",
    "zips",
    "ziti",
    "zoea",
    "zone",
    "zoon",
)  # noqa: E501, Q000, N806

ADJECTIVES = (
    "about",
    "above",
    "abuzz",
    "acerb",
    "acold",
    "acred",
    "added",
    "addle",
    "adept",
    "adult",
    "adunc",
    "adust",
    "afoul",
    "after",
    "agape",
    "agaze",
    "agile",
    "aging",
    "agley",
    "aglow",
    "ahead",
    "ahull",
    "aided",
    "alary",
    "algal",
    "alike",
    "alive",
    "alone",
    "aloof",
    "alpha",
    "amber",
    "amiss",
    "amort",
    "ample",
    "amuck",
    "angry",
    "anile",
    "apeak",
    "apish",
    "arced",
    "areal",
    "armed",
    "aroid",
    "ashen",
    "aspen",
    "astir",
    "atilt",
    "atrip",
    "aulic",
    "aural",
    "awash",
    "awful",
    "awing",
    "awned",
    "axile",
    "azoic",
    "azure",
    "baggy",
    "baked",
    "balky",
    "bally",
    "balmy",
    "banal",
    "bandy",
    "bardy",
    "bared",
    "barer",
    "barky",
    "basal",
    "based",
    "baser",
    "basic",
    "batty",
    "bawdy",
    "beady",
    "beaky",
    "beamy",
    "beaut",
    "beefy",
    "beery",
    "beige",
    "bendy",
    "bifid",
    "bijou",
    "biped",
    "birch",
    "bitty",
    "blame",
    "bland",
    "blank",
    "blear",
    "blest",
    "blind",
    "blond",
    "blown",
    "blowy",
    "bluer",
    "bluff",
    "blunt",
    "boned",
    "bonny",
    "boozy",
    "bored",
    "boric",
    "bosky",
    "bosom",
    "bound",
    "bovid",
    "bowed",
    "boxed",
    "braky",
    "brash",
    "brief",
    "briny",
    "brisk",
    "broad",
    "broch",
    "brood",
    "brown",
    "brute",
    "buggy",
    "bulgy",
    "bumpy",
    "burly",
    "burnt",
    "burry",
    "bushy",
    "busty",
    "butch",
    "buxom",
    "cadgy",
    "cagey",
    "calmy",
    "campy",
    "canny",
    "caped",
    "cased",
    "catty",
    "cauld",
    "cedar",
    "cered",
    "ceric",
    "chary",
    "cheap",
    "cheek",
    "chewy",
    "chief",
    "chill",
    "chirk",
    "choky",
    "cissy",
    "civil",
    "cleft",
    "coaly",
    "color",
    "comfy",
    "comic",
    "compo",
    "conic",
    "couth",
    "coxal",
    "crack",
    "crank",
    "crash",
    "crass",
    "crisp",
    "cronk",
    "cross",
    "crude",
    "cruel",
    "crumb",
    "cured",
    "curly",
    "curst",
    "cushy",
    "cutty",
    "cynic",
    "dated",
    "dazed",
    "dedal",
    "deism",
    "diazo",
    "dicey",
    "dingy",
    "direr",
    "dirty",
    "dishy",
    "dizzy",
    "dolce",
    "doped",
    "dopey",
    "dormy",
    "dorty",
    "dosed",
    "dotal",
    "dotty",
    "dowdy",
    "dowie",
    "downy",
    "dozen",
    "drawn",
    "dread",
    "drear",
    "dress",
    "dried",
    "ducky",
    "duddy",
    "dummy",
    "dumpy",
    "duple",
    "dural",
    "dusky",
    "dusty",
    "dutch",
    "dying",
    "eager",
    "eaten",
    "ebony",
    "edged",
    "eerie",
    "eight",
    "elder",
    "elect",
    "elfin",
    "elite",
    "empty",
    "enate",
    "enemy",
    "epoxy",
    "erect",
    "ethic",
    "every",
    "extra",
    "faced",
    "faery",
    "faint",
    "famed",
    "fancy",
    "farci",
    "fatal",
    "fated",
    "fatty",
    "fazed",
    "felon",
    "fenny",
    "ferny",
    "fetal",
    "fetid",
    "fewer",
    "fiery",
    "fifty",
    "filar",
    "filmy",
    "final",
    "fined",
    "finer",
    "finny",
    "fired",
    "first",
    "fishy",
    "fixed",
    "fizzy",
    "flaky",
    "flamy",
    "flash",
    "flawy",
    "fleet",
    "flory",
    "flown",
    "fluid",
    "fluky",
    "flush",
    "focal",
    "foggy",
    "folio",
    "forky",
    "forte",
    "forty",
    "found",
    "frail",
    "frank",
    "freed",
    "freer",
    "fresh",
    "fried",
    "front",
    "frore",
    "fuggy",
    "funky",
    "funny",
    "furry",
    "fusil",
    "fussy",
    "fuzzy",
    "gabby",
    "gamer",
    "gamey",
    "gamic",
    "gammy",
    "garni",
    "gauge",
    "gaunt",
    "gauzy",
    "gawky",
    "gawsy",
    "gemmy",
    "genal",
    "genic",
    "ghast",
    "gimpy",
    "girly",
    "glare",
    "glary",
    "glial",
    "glued",
    "gluey",
    "godly",
    "gooey",
    "goofy",
    "goosy",
    "gouty",
    "grade",
    "grand",
    "grapy",
    "grave",
    "gross",
    "group",
    "gruff",
    "guest",
    "gules",
    "gulfy",
    "gummy",
    "gushy",
    "gusty",
    "gutsy",
    "gutta",
    "gypsy",
    "gyral",
    "hadal",
    "hammy",
    "handy",
    "hardy",
    "hasty",
    "hated",
    "hazel",
    "heady",
    "heapy",
    "hefty",
    "heigh",
    "hempy",
    "herby",
    "hexed",
    "hi-fi",
    "hilly",
    "hired",
    "holey",
    "honey",
    "hooly",
    "hoven",
    "huger",
    "hulky",
    "humid",
    "hunky",
    "hyoid",
    "idled",
    "iliac",
    "inane",
    "incog",
    "inert",
    "inner",
    "inter",
    "iodic",
    "ionic",
    "irate",
    "irony",
    "itchy",
    "jaggy",
    "jammy",
    "japan",
    "jazzy",
    "jerky",
    "jetty",
    "joint",
    "jowly",
    "juicy",
    "jumpy",
    "jural",
    "kacha",
    "kaput",
    "kempt",
    "keyed",
    "kinky",
    "known",
    "kooky",
    "kraal",
    "laced",
    "laigh",
    "lairy",
    "lamer",
    "lardy",
    "larky",
    "lated",
    "later",
    "lathy",
    "leady",
    "leafy",
    "leaky",
    "leary",
    "least",
    "ledgy",
    "leery",
    "legal",
    "leggy",
    "lento",
    "level",
    "licht",
    "licit",
    "liege",
    "light",
    "liked",
    "liney",
    "lippy",
    "lived",
    "livid",
    "loamy",
    "loath",
    "lobar",
    "local",
    "loony",
    "loose",
    "loral",
    "losel",
    "lousy",
    "loved",
    "lower",
    "lowly",
    "lowse",
    "loyal",
    "lucid",
    "lucky",
    "lumpy",
    "lunar",
    "lurid",
    "lushy",
    "lying",
    "lyric",
    "macho",
    "macro",
    "magic",
    "major",
    "malar",
    "mangy",
    "manky",
    "manly",
    "mardy",
    "massy",
    "mated",
    "matte",
    "mauve",
    "mazed",
    "mealy",
    "meaty",
    "medal",
    "melic",
    "mesic",
    "mesne",
    "messy",
    "metal",
    "miffy",
    "milky",
    "mined",
    "minim",
    "minor",
    "minus",
    "mired",
    "mirky",
    "misty",
    "mixed",
    "modal",
    "model",
    "moire",
    "molar",
    "moldy",
    "moody",
    "moony",
    "mopey",
    "moral",
    "mossy",
    "mothy",
    "motor",
    "mousy",
    "moved",
    "mucid",
    "mucky",
    "muddy",
    "muggy",
    "muley",
    "mural",
    "murky",
    "mushy",
    "muted",
    "muzzy",
    "myoid",
    "naggy",
    "naive",
    "naked",
    "named",
    "nasty",
    "natal",
    "naval",
    "nervy",
    "newsy",
    "nicer",
    "niffy",
    "nifty",
    "ninth",
    "nitty",
    "nival",
    "noble",
    "nodal",
    "noisy",
    "non-U",
    "north",
    "nosed",
    "noted",
    "nowed",
    "nubby",
    "oaken",
    "oared",
    "oaten",
    "obese",
    "ocher",
    "ochre",
    "often",
    "ohmic",
    "oiled",
    "olden",
    "older",
    "oleic",
    "olive",
    "optic",
    "ortho",
    "osmic",
    "other",
    "outer",
    "ovoid",
    "owing",
    "owned",
    "paced",
    "pagan",
    "paled",
    "paler",
    "pally",
    "paper",
    "pappy",
    "parky",
    "party",
    "pasty",
    "pavid",
    "pawky",
    "peaky",
    "pearl",
    "peart",
    "peaty",
    "pedal",
    "peppy",
    "perdu",
    "perky",
    "pesky",
    "phony",
    "piano",
    "picky",
    "piled",
    "piney",
    "pious",
    "pique",
    "pithy",
    "platy",
    "plump",
    "plush",
    "podgy",
    "potty",
    "power",
    "prest",
    "pricy",
    "prima",
    "prime",
    "print",
    "privy",
    "prize",
    "prone",
    "proof",
    "prosy",
    "proud",
    "proxy",
    "pseud",
    "pucka",
    "pudgy",
    "puffy",
    "pukka",
    "pupal",
    "purer",
    "pursy",
    "pushy",
    "pyoid",
    "quack",
    "quare",
    "quasi",
    "quiet",
    "quits",
    "rabic",
    "rabid",
    "radio",
    "raked",
    "randy",
    "rapid",
    "rarer",
    "raspy",
    "rathe",
    "ratty",
    "ready",
    "reedy",
    "reeky",
    "refer",
    "regal",
    "riant",
    "ridgy",
    "right",
    "riled",
    "rimed",
    "rindy",
    "risen",
    "risky",
    "ritzy",
    "rival",
    "riven",
    "robed",
    "rocky",
    "roily",
    "roman",
    "rooky",
    "ropey",
    "round",
    "rowdy",
    "ruddy",
    "ruled",
    "rummy",
    "runic",
    "runny",
    "runty",
    "rural",
    "rusty",
    "rutty",
    "sable",
    "salic",
    "sandy",
    "sappy",
    "sarky",
    "sassy",
    "sated",
    "saved",
    "savvy",
    "scald",
    "scaly",
    "scary",
    "score",
    "scrap",
    "sedgy",
    "seely",
    "seral",
    "sewed",
    "shaky",
    "sharp",
    "sheen",
    "shier",
    "shill",
    "shoal",
    "shock",
    "shoed",
    "shore",
    "short",
    "shyer",
    "silky",
    "silly",
    "silty",
    "sixth",
    "sixty",
    "skint",
    "slack",
    "slant",
    "sleek",
    "slier",
    "slimy",
    "slung",
    "small",
    "smart",
    "smoky",
    "snaky",
    "sneak",
    "snide",
    "snowy",
    "snuff",
    "so-so",
    "soapy",
    "sober",
    "socko",
    "solar",
    "soled",
    "solid",
    "sonic",
    "sooth",
    "sooty",
    "soppy",
    "sorer",
    "sound",
    "soupy",
    "spent",
    "spicy",
    "spiky",
    "spiny",
    "spiry",
    "splay",
    "split",
    "sport",
    "spumy",
    "squat",
    "staid",
    "stiff",
    "still",
    "stoic",
    "stone",
    "stony",
    "store",
    "stout",
    "straw",
    "stray",
    "strip",
    "stung",
    "suave",
    "sudsy",
    "sulfa",
    "sulky",
    "sunny",
    "super",
    "sural",
    "surer",
    "surfy",
    "surgy",
    "surly",
    "swell",
    "swept",
    "swish",
    "sworn",
    "tabby",
    "taboo",
    "tacit",
    "tacky",
    "tamed",
    "tamer",
    "tangy",
    "taped",
    "tarot",
    "tarry",
    "tasty",
    "tatty",
    "taunt",
    "tawie",
    "teary",
    "techy",
    "telic",
    "tenor",
    "tense",
    "tenth",
    "tenty",
    "tepid",
    "terse",
    "testy",
    "third",
    "tidal",
    "tight",
    "tiled",
    "timid",
    "tinct",
    "tined",
    "tippy",
    "tipsy",
    "tonal",
    "toned",
    "tonic",
    "toric",
    "total",
    "tough",
    "toxic",
    "trade",
    "treed",
    "treen",
    "trial",
    "truer",
    "tubal",
    "tubby",
    "tumid",
    "tuned",
    "tutti",
    "twill",
    "typal",
    "typed",
    "typic",
    "umber",
    "unapt",
    "unbid",
    "uncut",
    "undue",
    "undug",
    "unfed",
    "unfit",
    "union",
    "unlet",
    "unmet",
    "unwed",
    "unwet",
    "upper",
    "upset",
    "urban",
    "utile",
    "uveal",
    "vagal",
    "valid",
    "vapid",
    "varus",
    "vatic",
    "veiny",
    "vital",
    "vivid",
    "vocal",
    "vogie",
    "volar",
    "vying",
    "wacky",
    "wally",
    "waney",
    "warty",
    "washy",
    "waspy",
    "waste",
    "waugh",
    "waxen",
    "webby",
    "wedgy",
    "weeny",
    "weepy",
    "weest",
    "weird",
    "welsh",
    "wersh",
    "whist",
    "white",
    "whity",
    "whole",
    "wider",
    "wight",
    "winey",
    "wired",
    "wised",
    "wiser",
    "withy",
    "wonky",
    "woods",
    "woozy",
    "world",
    "wormy",
    "worse",
    "worst",
    "woven",
    "wrath",
    "wrier",
    "wrong",
    "wroth",
    "xeric",
    "yarer",
    "yolky",
    "young",
    "yucky",
    "yummy",
    "zesty",
    "zingy",
    "zinky",
    "zippy",
    "zonal",
)  # noqa: E501, Q000, N806



from typing import Callable, Optional

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def answer_context_faithfulness_binary_factory(
    question_field: Optional[str] = "question", context_field: Optional[str] = "context", model: Optional[str] = "gpt-3.5-turbo-16k", is_azure: Optional[bool] = False
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that classifies if the generated answer was faithful to the given context.
    It is based on the paper [Evaluating Correctness and Faithfulness of Instruction-Following Models for Question Answering](https://arxiv.org/abs/2307.16877)
    which suggests using an LLM to flag any information in the generated answer that cannot be deduced from the given context.
    They find that GPT-4 is the best model for this analysis as measured by correlation with human judgment.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        question_field: The key name/field used for the question/query of the user. Defaults to "question".
        context_field: The key name/field used for the retrieved context. Defaults to "context".
        model: The model which should be used for grading. Currently, only supports OpenAI chat models. Defaults to "gpt-4".

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        if the generated answer was faithful to the given context.
    """

    def answer_context_faithfulness_binary(log: Log) -> float:
        question = log.inputs[question_field]
        evidence = log.inputs[context_field]
        output = log.output
        response = call_openai(
            model=model,
            messages=[
                {"role": "system", "content": "You are CompareGPT, a machine to verify the groundedness of predictions. Answer with " "only yes/no."},
                {
                    "role": "user",
                    "content": f"You are given a question, the corresponding evidence and a prediction from a model. Compare "
                    f'the "Prediction" and the "Evidence" to determine whether all the information of the '
                    f"prediction in present in the evidence or can be inferred from the evidence. You must answer "
                    f'"no" if there are any specific details in the prediction that are not mentioned in the '
                    f"evidence or cannot be inferred from the evidence.\n\n"
                    f"Question: {question}\n\nPrediction: {output}\n\nEvidence: {evidence}\n\nCompareGPT response:",
                },
            ],
            temperature=0.0,
            is_azure=is_azure,
        )
        return float("yes" in response.lower())

    return answer_context_faithfulness_binary



from typing import Any, Dict, Optional

import json
import logging
import os

from attrs import asdict, define, field
from cattrs import structure

from parea.api_client import HTTPClient
from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.helpers import serialize_metadata_values
from parea.schemas.log import TraceIntegrations
from parea.schemas.models import CreateGetProjectResponseSchema, TraceLog, UpdateLog
from parea.utils.trace_integrations.langchain_utils import _dumps_json
from parea.utils.universal_encoder import json_dumps

logger = logging.getLogger()

LOG_ENDPOINT = "/trace_log"
VENDOR_LOG_ENDPOINT = "/trace_log/{vendor}"


@define
class PareaLogger:
    _client: HTTPClient = field(init=False, default=None)
    _project_uuid: str = field(init=False, default=None)
    _project_name: str = field(init=False, default=None)

    def set_client(self, client: HTTPClient) -> None:
        self._client = client

    def set_project_uuid(self, project_uuid: str, project_name: str) -> None:
        self._project_uuid = project_uuid
        self._project_name = project_name

    def _get_project_uuid(self) -> str:
        try:
            if not self._project_uuid:
                self._project_uuid = self._create_or_get_project(self._project_name or "default").uuid
            return self._project_uuid
        except Exception as e:
            logger.error(f"PareaLogger: Error getting project uuid for project {self._project_name}: {e}")
            raise

    def _create_or_get_project(self, name: str) -> CreateGetProjectResponseSchema:
        r = self._client.request(
            "POST",
            "/project",
            data={"name": name},
        )
        return structure(r.json(), CreateGetProjectResponseSchema)

    def update_log(self, data: UpdateLog) -> None:
        data = serialize_metadata_values(data)
        self._client.request(
            "PUT",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    def record_log(self, data: TraceLog) -> None:
        data = serialize_metadata_values(data)
        data.project_uuid = self._get_project_uuid()
        self._client.request(
            "POST",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    async def arecord_log(self, data: TraceLog) -> None:
        data = serialize_metadata_values(data)
        data.project_uuid = self._get_project_uuid()
        await self._client.request_async(
            "POST",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    def default_log(self, data: TraceLog) -> None:
        if self._client:
            if data.target:
                data.target = json_dumps(data.target)
            self.record_log(data)

    def record_vendor_log(self, data: Dict[str, Any], vendor: TraceIntegrations) -> None:
        data["project_uuid"] = self._get_project_uuid()
        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data["experiment_uuid"] = experiment_uuid
        self._client.add_integration("langchain")
        self._client.request(
            "POST",
            VENDOR_LOG_ENDPOINT.format(vendor=vendor.value),
            data=json.loads(_dumps_json(data)),  # uuid is not serializable
        )

    async def arecord_vendor_log(self, data: Dict[str, Any], vendor: TraceIntegrations) -> None:
        data["project_uuid"] = self._get_project_uuid()
        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data["experiment_uuid"] = experiment_uuid
        self._client.add_integration("langchain")
        await self._client.request_async(
            "POST",
            VENDOR_LOG_ENDPOINT.format(vendor=vendor.value),
            data=json.loads(_dumps_json(data)),  # uuid is not serializable
        )


parea_logger = PareaLogger()



from typing import Callable, List, Optional

from parea.evals.utils import call_openai, get_context, sent_tokenize
from parea.schemas.log import Log


def context_query_relevancy_factory(
    question_field: str = "question", context_fields: Optional[List[str]] = None, model: Optional[str] = "gpt-3.5-turbo-16k", is_azure: Optional[bool] = False
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures how relevant the retrieved context is to the given question.
    It is based on the paper [RAGAS: Automated Evaluation of Retrieval Augmented Generation](https://arxiv.org/abs/2309.15217)
    which suggests using an LLM to extract any sentence from the retrieved context relevant to the query. Then, calculate
    the ratio of relevant sentences to the total number of sentences in the retrieved context.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        model: The model which should be used for grading. Defaults to "gpt-3.5-turbo-16k".
        question_field: The key name/field used for the question/query of the user. Defaults to "question".
        context_fields: An optional list of key names/fields used for the retrieved contexts in the input to function. If empty list or None, it will use the output field of the log as context. Defaults to None.

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        if the retrieved context is relevant to the query.
    """

    def context_query_relevancy(log: Log) -> float:
        """Quantifies how much the retrieved context relates to the query."""
        question = log.inputs[question_field]
        context = get_context(log, context_fields)

        extracted_sentences = call_openai(
            model=model,
            messages=[
                {
                    "role": "user",
                    "content": f"""\
Please extract relevant sentences from the provided context that is absolutely required answer the following question. If no relevant sentences are found, or if you believe the question cannot be answered from the given context, return the phrase "Insufficient Information".  While extracting candidate sentences you're not allowed to make any changes to sentences from given context.

question:{question}
context:\n{context}
candidate sentences:\n""",
                }
            ],
            temperature=0.0,
            is_azure=is_azure,
        ).strip()
        if "insufficient information" in extracted_sentences.lower() and abs(len(extracted_sentences) - len("insufficient information")) < 10:
            return 0.0
        else:
            n_extracted_sentences = len(sent_tokenize(extracted_sentences))
            n_context_sentences = len(sent_tokenize(context))
            return n_extracted_sentences / n_context_sentences

    return context_query_relevancy



from typing import List, Optional

from attr import asdict

from parea.cache.cache import Cache
from parea.schemas.models import CacheRequest, TraceLog
from parea.utils.universal_encoder import json_dumps


class InMemoryCache(Cache):
    def __init__(self):
        self.cache = {}
        self.logs = []

    def get(self, key: CacheRequest) -> Optional[TraceLog]:
        return self.cache.get(json_dumps(asdict(key)))

    async def aget(self, key: CacheRequest) -> Optional[TraceLog]:
        return self.get(key)

    def set(self, key: CacheRequest, value: TraceLog):
        self.cache[json_dumps(asdict(key))] = value

    async def aset(self, key: CacheRequest, value: TraceLog):
        self.set(key, value)

    def invalidate(self, key: CacheRequest):
        key = json_dumps(asdict(key))
        if key in self.cache:
            del self.cache[key]

    async def ainvalidate(self, key: CacheRequest):
        self.invalidate(key)

    def log(self, value: TraceLog):
        self.logs.append(value)

    def read_logs(self) -> List[TraceLog]:
        return self.logs.copy()



from typing import Callable, Optional

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def lm_vs_lm_factuality_factory(
    examiner_model: str = "gpt-4",
    is_azure: Optional[bool] = False,
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures the factuality of an LLM's response to a given question.
    It is based on the paper [LM vs LM: Detecting Factual Errors via Cross Examination](https://arxiv.org/abs/2305.13281) which proposes using
    another LLM to assess an LLM response's factuality. To do this, the examining LLM generates follow-up questions to the
    original response until it can confidently determine the factuality of the response.
    This method outperforms prompting techniques such as asking the original model, "Are you sure?" or instructing the
    model to say, "I don't know," if it is uncertain.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        examiner_model: The model which will examine the original model. Currently, only supports OpenAI chat models.

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        the factuality of the response.
    """

    def lm_vs_lm_factuality(log: Log) -> float:
        output = log.output
        messages_examinee = [m.to_dict() for m in log.configuration.messages]

        # ask examiner for follow-up questions
        setup_prompt = f"""Your goal is to try to verify the correctness of the following claim: "{output}", based on the background information you will gather. To gather this, You will provide short questions whose purpose will be to verify the correctness of the claim, and I will reply to you with the answers to these. Hopefully, with the help of the background questions and their answers, you will be able to reach a conclusion as to whether the claim is correct or possibly incorrect. Please keep asking questions as long as you’re yet to be sure regarding the true veracity of the claim. Please start with the first questions."""
        messages_examiner = [{"role": "user", "content": setup_prompt}]
        follow_up_questions = call_openai(
            model=examiner_model,
            messages=messages_examiner,
            temperature=0.0,
            is_azure=is_azure,
        )
        messages_examiner += [{"role": "assistant", "content": follow_up_questions}]
        n_rounds_follow_up_questions = 1

        follow_up_prompt = """(i) Do you have any follow-up questions? Please answer with Yes or No.
    (ii) What are the follow-up questions?"""
        # ask examinee follow-up questions until they reach a conclusion
        while follow_up_questions is not None:
            messages_examinee += [{"role": "user", "content": follow_up_questions}]
            follow_up_answers = call_openai(
                model=examiner_model if is_azure else log.configuration.model,
                messages=messages_examinee,
                temperature=log.configuration.model_params.temp,
                top_p=log.configuration.model_params.top_p,
                frequency_penalty=log.configuration.model_params.frequency_penalty,
                presence_penalty=log.configuration.model_params.presence_penalty,
                max_tokens=log.configuration.model_params.max_length,
                response_format=log.configuration.model_params.response_format,
                is_azure=is_azure,
            )
            messages_examiner.append({"role": "assistant", "content": follow_up_answers})

            if n_rounds_follow_up_questions > 3:
                break
            else:
                messages_examiner.append({"role": "user", "content": follow_up_prompt})
                n_rounds_follow_up_questions += 1

            examiner_response = call_openai(
                model=examiner_model,
                messages=messages_examiner,
                temperature=0.0,
                is_azure=is_azure,
            )
            messages_examiner += [{"role": "assistant", "content": examiner_response}]
            if "yes" in examiner_response.lower():
                follow_up_questions = examiner_response
                messages_examinee += [{"role": "assistant", "content": follow_up_answers}]
            else:
                follow_up_questions = None

        # ask examiner for their conclusion
        factuality_decision_prompt = (
            """Based on the interviewee’s answers to your questions, what is your conclusion regarding the correctness of the claim? Do you think it is correct or incorrect?"""
        )
        messages_examiner += [{"role": "user", "content": factuality_decision_prompt}]
        examiner_response = call_openai(
            model=examiner_model,
            messages=messages_examiner,
            temperature=0.0,
            is_azure=is_azure,
        )
        return float("incorrect" not in examiner_response.lower())

    return lm_vs_lm_factuality


lm_vs_lm_factuality_gpt4 = lm_vs_lm_factuality_factory("gpt-4")
lm_vs_lm_factuality_gpt3t = lm_vs_lm_factuality_factory("gpt-3.5-turbo-16k")



from typing import Union

from parea.evals.utils import call_openai, sent_tokenize
from parea.schemas.log import Log


def self_check(log: Log) -> Union[float, None]:
    """
    Given that many API-based LLMs don't reliably give access to the log probabilities of the generated tokens, assessing
    the certainty of LLM predictions via perplexity isn't possible.
    The [SelfCheckGPT: Zero-Resource Black-Box Hallucination Detection for Generative Large Language Models](https://arxiv.org/abs/2303.08896) paper
    suggests measuring the average factuality of every sentence in a generated response. They generate additional responses
    from the LLM at a high temperature and check how much every sentence in the original answer is supported by the other generations.
    The intuition behind this is that if the LLM knows a fact, it's more likely to sample it. The authors find that this
    works well in detecting non-factual and factual sentences and ranking passages in terms of factuality.
    The authors noted that correlation with human judgment doesn't increase after 4-6 additional
    generations when using `gpt-3.5-turbo` to evaluate biography generations.

    Args:
        log (Log): The log object to of the trace evaluate.

    Returns:
        float: A score between 0 and 1 indicating the factuality of the response.
    """
    if log.configuration is None or log.configuration.messages is None:
        return None

    messages = [m.to_dict() for m in log.configuration.messages]

    n_sampled_outputs = 5
    sampled_outputs = []
    for _ in range(n_sampled_outputs):
        response = call_openai(
            messages=messages,
            model=log.configuration.model,
            temperature=1.0,
            max_tokens=log.configuration.model_params.max_length,
            top_p=log.configuration.model_params.top_p,
            frequency_penalty=log.configuration.model_params.frequency_penalty,
            presence_penalty=log.configuration.model_params.presence_penalty,
            response_format=log.configuration.model_params.response_format,
        )
        sampled_outputs.append(response)

    sentences = sent_tokenize(log.output)

    if len(sentences) == 0:
        return 0.0

    sentences_scores = []
    for sentence in sentences:
        scores = []
        for sampled_output in sampled_outputs:
            response = call_openai(
                messages=[
                    {
                        "role": "user",
                        "content": f"""Context: {sampled_output}
Sentence: {sentence}
Is the sentence supported by the context above?
Answer Yes or No:""",
                    }
                ],
                model="gpt-3.5-turbo",
                temperature=0.0,
            )
            scores.append(float("yes" in response.lower()))
        sentences_scores.append(sum(scores) / len(scores))

    return sum(sentences_scores) / len(sentences_scores)



from typing import Callable, Dict, List, Optional, Union

import json
import re
import sys
from functools import lru_cache, wraps

import tiktoken
from openai import NotGiven
from openai import __version__ as openai_version
from pydantic._internal._model_construction import ModelMetaclass

from parea.constants import ALL_NON_AZURE_MODELS_INFO, AZURE_MODEL_INFO, TURN_OFF_PAREA_EVAL_LOGGING
from parea.parea_logger import parea_logger
from parea.schemas.log import LLMInputs, Message, ModelParams, Role
from parea.schemas.models import UpdateTraceScenario
from parea.utils.trace_utils import fill_trace_data, get_current_trace_id, log_in_thread, trace_data, trace_insert
from parea.utils.universal_encoder import json_dumps

is_openai_1 = openai_version.startswith("1.")
if is_openai_1:
    from openai.types.chat import ChatCompletion, ChatCompletionMessage


# https://gist.github.com/JettJones/c236494013f22723c1822126df944b12
def skip_decorator_if_func_in_stack_for_evals(*funcs_to_check: Callable) -> Callable:
    def decorator_wrapper(decorator: Callable) -> Callable:
        def new_decorator(self, func: Callable) -> Callable:  # Include self
            @wraps(func)
            def wrapper(*args, **kwargs):
                if not TURN_OFF_PAREA_EVAL_LOGGING:
                    return decorator(self, func)(*args, **kwargs)  # Include self
                frame = sys._getframe().f_back
                caller_names = ""
                while frame:
                    caller_names += frame.f_code.co_name + "|"
                    frame = frame.f_back
                if any(func_to_check.__name__ in caller_names for func_to_check in funcs_to_check):
                    return func(*args, **kwargs)
                return decorator(self, func)(*args, **kwargs)  # Include self

            return wrapper

        return new_decorator

    return decorator_wrapper


def _safe_encode(encoding, text):
    try:
        return len(encoding.encode(text))
    except Exception as e:
        print(f"Error encoding text: {e}")
        return 0


def _num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613", is_azure: bool = False):
    """Return the number of tokens used by a list of messages.
    source: https://cookbook.openai.com/examples/how_to_count_tokens_with_tiktoken
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if (
        model
        in {
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-1106",
            "gpt-3.5-turbo-0125",
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
            "gpt-4-turbo-preview",
            "gpt-4-1106-preview",
            "gpt-4-0125-preview",
        }
        or is_azure
    ):
        tokens_per_message = 3
        tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-4" in model:
        print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
        return _num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        print(
            f"""Warning: num_tokens_from_messages() is not implemented for model {model}.
            See https://github.com/openai/openai-python/blob/main/chatml.md for
            information on how messages are converted to tokens. Will use default model"""
        )
        return _num_tokens_from_messages(messages)
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += _safe_encode(encoding, value)
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def _num_tokens_from_functions(functions, function_call, model="gpt-3.5-turbo-0613"):
    """Return the number of tokens used by a list of functions. modified from source:
    https://community.openai.com/t/how-to-calculate-the-tokens-when-using-function-call/266573/10 tested:
    https://deepnote.com/workspace/joel-alexander-7514-38647c2d-ba7d-4573-b016-5233b373912e/project/BloomreachDemo
    -66683e80-4414-4b43-ac70-e7eabd3743c1/notebook/token_counting_openai-992f53bba3224221901dcb88befa2051
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 3 if model in ["gpt-3.5-turbo-1106", "gpt-4-1106-preview"] else 0
    for function in functions:
        try:
            function_tokens = _safe_encode(encoding, function.get("name", ""))
            function_tokens += _safe_encode(encoding, function.get("description", ""))
        except Exception as e:
            print(f"Error counting tokens: {e}")
            function_tokens = 0

        if "parameters" in function:
            parameters = function["parameters"]
            if "properties" in parameters:
                for propertiesKey in parameters["properties"]:
                    function_tokens += _safe_encode(encoding, propertiesKey)
                    v = parameters["properties"][propertiesKey]
                    for field in v:
                        if field == "type":
                            function_tokens += 2
                            function_tokens += _safe_encode(encoding, v["type"])
                        elif field == "description":
                            function_tokens += 2
                            function_tokens += _safe_encode(encoding, v["description"])
                        elif field == "enum":
                            function_tokens -= 3
                            for o in v["enum"]:
                                function_tokens += 3
                                function_tokens += _safe_encode(encoding, o)
                        else:
                            print(f"Warning: not supported field {field}")
                function_tokens += 11

        num_tokens += function_tokens

    num_tokens += 10
    function_call_tokens = min(_safe_encode(encoding, "auto") - 1, 0)
    if isinstance(function_call, dict):
        function_call_tokens = min(_safe_encode(encoding, json_dumps(function_call)) - 1, 0)
    return num_tokens + function_call_tokens


def _num_tokens_from_string(string: str, model_name: str = "gpt-3.5-turbo") -> int:
    """Returns the number of tokens in a text string."""
    try:
        encoding = tiktoken.encoding_for_model(model_name)
    except KeyError:
        print(f"Warning: model {model_name} not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    return _safe_encode(encoding, string)


def _calculate_input_tokens(
    messages: Optional[List[Dict[str, str]]],
    functions: List[Dict[str, str]],
    function_call: Union[str, Dict[str, str]],
    model: str,
) -> int:
    is_azure = model.startswith("azure_") or model in AZURE_MODEL_INFO
    num_function_tokens = _num_tokens_from_functions(functions, function_call, model)
    num_input_tokens = _num_tokens_from_string(json_dumps(messages), model) if model == "gpt-4-vision-preview" else _num_tokens_from_messages(messages, model, is_azure)
    return num_input_tokens + num_function_tokens


def _format_function_call(response_message) -> str:
    def clean_json_string(s):
        """If OpenAI responds with improper newlines and multiple quotes, this will clean it up"""
        return json_dumps(s.replace("'", '"').replace("\\n", "\\\\n"))

    func_obj = response_message.tool_calls
    if response_message.function_call and response_message.function_call.name:
        func_obj = response_message.function_call

    calls = []
    if not isinstance(func_obj, list):
        func_obj = [func_obj]

    for call in func_obj:
        if call:
            is_tool_call = hasattr(call, "id")
            body = getattr(call, "function", None) or call
            function_name = body.name
            try:
                function_args = json.loads(body.arguments)
            except json.decoder.JSONDecodeError:
                try:
                    function_args = json.loads(clean_json_string(body.arguments))
                except Exception:
                    function_args = str(body.arguments)

            if is_tool_call:
                calls.append(
                    {
                        "id": call.id,
                        "type": call.type,
                        "function": {"name": function_name, "arguments": function_args},
                    }
                )
            else:
                calls.append({"name": function_name, "arguments": function_args})
    return json_dumps(calls, indent=4)


def _resolve_functions(kwargs):
    if "functions" in kwargs:
        f = kwargs.get("functions", [])
        return None if isinstance(f, NotGiven) else f
    elif "tools" in kwargs:
        tools = kwargs["tools"]
        if isinstance(tools, NotGiven):
            return None
        if isinstance(tools, list):
            return [d.get("function", {}) for d in tools]

        return []  # it is either a list or Stainless's `NotGiven`

    return []


def _kwargs_to_llm_configuration(kwargs, model=None) -> LLMInputs:
    functions = _resolve_functions(kwargs)
    function_call_default = "auto" if functions else None
    function_call = kwargs.get("function_call", function_call_default) or kwargs.get("tool_choice", function_call_default)
    response_format = kwargs.get("response_format", None)
    response_format = {"type": "json_schema", "json_schema": str(response_format)} if isinstance(response_format, ModelMetaclass) else response_format
    temp = kwargs.get("temperature", 1.0)
    max_length = kwargs.get("max_tokens", None)
    top_p = kwargs.get("top_p", 1.0)
    frequency_penalty = kwargs.get("frequency_penalty", 0.0)
    presence_penalty = kwargs.get("presence_penalty", 0.0)
    return LLMInputs(
        model=model or kwargs.get("model", None),
        provider="openai",
        messages=_convert_oai_messages(kwargs.get("messages", None)),
        functions=functions,
        function_call=None if isinstance(function_call, NotGiven) else function_call,
        model_params=ModelParams(
            temp=None if isinstance(temp, NotGiven) else temp,
            max_length=None if isinstance(max_length, NotGiven) else max_length,
            top_p=None if isinstance(top_p, NotGiven) else top_p,
            frequency_penalty=None if isinstance(frequency_penalty, NotGiven) else frequency_penalty,
            presence_penalty=None if isinstance(presence_penalty, NotGiven) else presence_penalty,
            response_format=response_format,
        ),
    )


def _convert_oai_messages(messages: list) -> Union[List[Union[dict, Message]], None]:
    if not messages:
        return messages
    if is_openai_1:
        cleaned_messages = []
        for m in messages:
            is_chat_completion = isinstance(m, ChatCompletionMessage)
            if (is_chat_completion and m.role == "tool") or (isinstance(m, dict) and m.get("role") == "tool"):
                tool_call_id = m.tool_call_id if is_chat_completion else m.get("tool_call_id")
                content = m.content if is_chat_completion else m.get("content", "")
                cleaned_messages.append(
                    Message(
                        role=Role.tool,
                        content=json_dumps({"tool_call_id": tool_call_id, "content": content}, indent=4),
                    )
                )
            elif is_chat_completion:
                cleaned_messages.append(
                    Message(
                        role=m.role,
                        content=m.content if m.content else _format_function_call(m),
                    )
                )
            else:
                if "tool_calls" in m:
                    cleaned_messages.append(
                        Message(
                            role=m["role"],
                            content=_format_function_call(ChatCompletionMessage(**m)),
                        )
                    )
                else:
                    cleaned_messages.append(m)
        return cleaned_messages
    else:
        return messages


@lru_cache(maxsize=128)
def _compute_cost(prompt_tokens: int, completion_tokens: int, model: str) -> float:
    if model in AZURE_MODEL_INFO:
        cost_per_token = AZURE_MODEL_INFO[model]
    else:
        cost_per_token = ALL_NON_AZURE_MODELS_INFO.get(model, {"prompt": 0, "completion": 0})
    cost = ((prompt_tokens * cost_per_token["prompt"]) + (completion_tokens * cost_per_token["completion"])) / 1_000_000
    cost = round(cost, 10)
    return cost


def _process_response(response, model_inputs, trace_id):
    response_message = response.choices[0].message
    if response_message.finish_reason == "content_filter":
        trace_insert({"error": "Error: The content was filtered due to policy violations."}, trace_id)
    if hasattr(response_message, "refusal"):
        completion = response_message.refusal
    elif response_message.content:
        completion = response_message.content
    else:
        completion = _format_function_call(response_message)

    usage = response.usage
    trace_insert(
        {
            "configuration": _kwargs_to_llm_configuration(model_inputs, response.model),
            "output": completion,
            "input_tokens": usage.prompt_tokens,
            "output_tokens": usage.completion_tokens,
            "total_tokens": usage.prompt_tokens + usage.completion_tokens,
            "cost": _compute_cost(usage.prompt_tokens, usage.completion_tokens, response.model),
        },
        trace_id,
    )


def _process_stream_response(content: list, tools: dict, data: dict, trace_id: str):
    # format to Log schema and send to logger
    model = data.get("model")
    final_content = "".join(content)

    for tool in tools.values():
        tool["function"]["arguments"] = "".join(tool["function"]["arguments"])

    tool_calls = [t["function"] for t in tools.values()]
    for tool in tool_calls:
        try:
            tool["arguments"] = json.loads(tool["arguments"])
        except Exception:
            tool["arguments"] = str(tool["arguments"])

    completion = final_content or json_dumps(tool_calls, indent=4)

    prompt_tokens = _calculate_input_tokens(
        data.get("messages", []),
        data.get("functions", []) or [d["function"] for d in data.get("tools", [])],
        data.get("function_call", "auto") or data.get("tool_choice", "auto"),
        data.get("model"),
    )
    completion_tokens = _num_tokens_from_string(final_content if final_content else json_dumps(tool_calls), model)
    data = {
        "configuration": _kwargs_to_llm_configuration(data, model),
        "output": completion,
        "input_tokens": prompt_tokens,
        "output_tokens": completion_tokens,
        "total_tokens": prompt_tokens + completion_tokens,
        "cost": _compute_cost(prompt_tokens, completion_tokens, model),
    }
    fill_trace_data(trace_id, data, UpdateTraceScenario.OPENAICONFIG)
    data_with_config = trace_data.get()[trace_id]
    parea_logger.default_log(data=data_with_config)


def convert_openai_raw_stream_to_log(content: list, tools: dict, data: dict, trace_id: str):
    log_in_thread(
        _process_stream_response,
        {"content": content, "tools": tools, "data": data, "trace_id": trace_id},
    )


def convert_openai_raw_to_log(r: dict, data: dict):
    log_in_thread(
        _process_response,
        {
            "response": ChatCompletion(**r),
            "model_inputs": data,
            "trace_id": get_current_trace_id(),
        },
    )


def safe_format_template_to_prompt(_template: str, **kwargs) -> str:
    """Replaces langchain.prompts.PromptTemplate.format in a safe manner.

    Only variables '{{...}}' will be replaced, not any in '{...}'
    """

    def replace(match):
        var_name = match.group(1)
        return str(kwargs.get(var_name, match.group(0)))

    return re.sub(r"{{(\w+)}}", replace, _template)



from typing import Callable, List, Optional

from parea.evals.utils import call_openai, get_context, safe_json_loads
from parea.schemas.log import Log


def context_ranking_pointwise_factory(
    question_field: str = "question",
    context_fields: Optional[List[str]] = None,
    ranking_measurement="average_precision",
    model: Optional[str] = "gpt-3.5-turbo-16k",
    is_azure: Optional[bool] = False,
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures how well the retrieved contexts are ranked by relevancy to the given query
    by pointwise estimation of the relevancy of every context to the query. It is based on the paper
    [RAGAS: Automated Evaluation of Retrieval Augmented Generation](https://arxiv.org/abs/2309.15217) which suggests using an LLM
    to check if every extracted context is relevant. Then, they measure how well the contexts are ranked by calculating the
    mean average precision. Note that this approach considers any two relevant contexts equally important/relevant to the query.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        model: The model which should be used for grading. Defaults to "gpt-3.5-turbo-16k".
        question_field: The key name/field used for the question/query of the user. Defaults to "question".
        context_fields: An optional list of key names/fields used for the retrieved contexts in the input to function. If empty list or None, it will use the output field of the log as context. Defaults to None.
        ranking_measurement: Method to calculate ranking. Currently, only supports "average_precision".

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        how well the retrieved context is ranked by their relevancy.

    Raises:
        ImportError: If numpy is not installed.
    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Please install numpy to use this metric.")

    def context_ranking_pointwise(log: Log) -> float:
        """Quantifies if the retrieved context is ranked by their relevancy"""
        question = log.inputs[question_field]
        contexts = get_context(log, context_fields, True)

        verifications = []
        for context in contexts:
            response = call_openai(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": f"""\
Verify if the information in the given context is useful in answering the question.

question: What are the health benefits of green tea?
context:
This article explores the rich history of tea cultivation in China, tracing its roots back to the ancient dynasties. It discusses how different regions have developed their unique tea varieties and brewing techniques. The article also delves into the cultural significance of tea in Chinese society and how it has become a symbol of hospitality and relaxation.
verification:
{{"reason":"The context, while informative about the history and cultural significance of tea in China, does not provide specific information about the health benefits of green tea. Thus, it is not useful for answering the question about health benefits.", "verdict":"No"}}

question: How does photosynthesis work in plants?
context:
Photosynthesis in plants is a complex process involving multiple steps. This paper details how chlorophyll within the chloroplasts absorbs sunlight, which then drives the chemical reaction converting carbon dioxide and water into glucose and oxygen. It explains the role of light and dark reactions and how ATP and NADPH are produced during these processes.
verification:
{{"reason":"This context is extremely relevant and useful for answering the question. It directly addresses the mechanisms of photosynthesis, explaining the key components and processes involved.", "verdict":"Yes"}}

question:{question}
context:
{context}
verification:""",
                    }
                ],
                temperature=0.0,
                is_azure=is_azure,
            )
            verifications.append(response)

        if ranking_measurement == "average_precision":
            response = [safe_json_loads(item) for item in verifications]
            response = [int("yes" in resp.get("verdict", " ").lower()) if resp.get("verdict") else np.nan for resp in response]
            denominator = sum(response) + 1e-10
            numerator = sum([(sum(response[: i + 1]) / (i + 1)) * response[i] for i in range(len(response))])
            return numerator / denominator
        else:
            raise NotImplementedError

    return context_ranking_pointwise



from typing import Callable, List, Optional, Union

import re

from parea.evals.utils import call_openai, get_context
from parea.schemas.log import Log


def percent_target_supported_by_context_factory(
    question_field: str = "question", context_fields: Optional[List[str]] = None, model: Optional[str] = "gpt-3.5-turbo-16k", is_azure: Optional[bool] = False
) -> Callable[[Log], Union[float, None]]:
    """Quantifies how many sentences in the target/ground truth are supported by the retrieved context."""

    def percent_target_supported_by_context(log: Log) -> Union[float, None]:
        """Quantifies how many sentences in the target/ground truth are supported by the retrieved context."""
        question = log.inputs[question_field]
        context = get_context(log, context_fields)

        if (target := log.target) is None:
            return None

        classification = call_openai(
            model=model,
            messages=[
                {
                    "role": "user",
                    "content": f"""\
Given a context, and an answer, analyze each sentence in the answer and classify if the sentence can be attributed to the given context or not. Output json with reason.


question: What can you tell me about albert Albert Einstein?
context: Albert Einstein (14 March 1879 – 18 April 1955) was a German-born theoretical physicist,widely held to be one of the greatest and most influential scientists of all time. Best known for developing the theory of relativity, he also made important contributions to quantum mechanics, and was thus a central figure in the revolutionary reshaping of the scientific understanding of nature that modern physics accomplished in the first decades of the twentieth century. His mass–energy equivalence formula E = mc2, which arises from relativity theory, has been called "the world's most famous equation". He received the 1921 Nobel Prize in Physics "for his services to theoretical physics, and especially for his discovery of the law of the photoelectric effect", a pivotal step in the development of quantum theory. His work is also known for its influence on the philosophy of science. In a 1999 poll of 130 leading physicists worldwide by the British journal Physics World, Einstein was ranked the greatest physicist of all time. His intellectual achievements and originality have made Einstein synonymous with genius.
answer: Albert Einstein born in 14 March 1879 was  German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time. He received the 1921 Nobel Prize in Physics "for his services to theoretical physics. He published 4 papers in 1905.  Einstein moved to Switzerland in 1895
classification:
[
    {{
        "statement_1":"Albert Einstein, born on 14 March 1879, was a German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time.",
        "reason": "The date of birth of Einstein is mentioned clearly in the context.",
        "Attributed": "Yes"
    }},
    {{
        "statement_2":"He received the 1921 Nobel Prize in Physics 'for his services to theoretical physics.",
        "reason": "The exact sentence is present in the given context.",
        "Attributed": "Yes"
    }},
    {{
        "statement_3": "He published 4 papers in 1905.",
        "reason": "There is no mention about papers he wrote in the given context.",
        "Attributed": "No"
    }},
    {{
        "statement_4":"Einstein moved to Switzerland in 1895.",
        "reason": "There is no supporting evidence for this in the given context.",
        "Attributed": "No"
    }}
]

question: who won 2020 icc world cup?
context: Who won the 2022 ICC Men's T20 World Cup?
The 2022 ICC Men's T20 World Cup, held from October 16 to November 13, 2022, in Australia, was the eighth edition of the tournament. Originally scheduled for 2020, it was postponed due to the COVID-19 pandemic. England emerged victorious, defeating Pakistan by five wickets in the final to clinch their second ICC Men's T20 World Cup title.
answer: England
classification:
[
    {{
        "statement_1":"England won the 2022 ICC Men's T20 World Cup.",
        "reason": "From context it is clear that England defeated Pakistan to win the World Cup.",
         "Attributed": "Yes"
    }}
]

question: {question}
context: {context}
answer: {target}
classification:
""",
                }
            ],
            temperature=0.0,
            is_azure=is_azure,
        )
        pattern = r"\[\s*\{.*?\}(\s*,\s*\{.*?\})*\s*\]"
        match = re.search(pattern, classification.replace("\n", ""))
        if match:
            response = eval(classification)
            numerator = sum(item.get("Attributed").lower() == "yes" for item in response)
            return numerator / len(response)
        else:
            return 0.0

    return percent_target_supported_by_context



from typing import Any, Callable, List, Mapping, Tuple

import contextvars
import logging
from json import JSONDecodeError

from instructor.retry import InstructorRetryException
from pydantic import ValidationError
from wrapt import wrap_object

from parea import trace
from parea.helpers import gen_trace_id
from parea.schemas import EvaluationResult, UpdateLog
from parea.utils.trace_integrations.wrapt_utils import CopyableFunctionWrapper
from parea.utils.trace_utils import logger_update_record, trace_data, trace_insert
from parea.utils.universal_encoder import json_dumps

logger = logging.getLogger()

instructor_trace_id = contextvars.ContextVar("instructor_trace_id", default="")
instructor_val_err_count = contextvars.ContextVar("instructor_val_err_count", default=0)
instructor_val_errs = contextvars.ContextVar("instructor_val_errs", default=[])


def instrument_instructor_validation_errors() -> None:
    for retry_method in ["retry_async", "retry_sync"]:
        wrap_object(
            module="instructor.patch",
            name=f"{retry_method}",
            factory=CopyableFunctionWrapper,
            args=(_RetryWrapper(),),
        )

    wrap_object(
        module="tenacity",
        name="AttemptManager.__exit__",
        factory=CopyableFunctionWrapper,
        args=(_AttemptManagerExitWrapper(),),
    )


def get_reasons(exception: Exception) -> list[str]:
    if isinstance(exception, InstructorRetryException):
        return [str(arg) for arg in exception.args]
    return [str(exception)]


def report_instructor_validation_errors() -> None:
    reason = "\n\n\n".join(instructor_val_errs.get())
    if reason:
        reason = "\n" + reason
    instructor_score = EvaluationResult(
        name="instructor_validation_error_count",
        score=instructor_val_err_count.get(),
        reason=reason,
    )
    trace_update_dict = {"scores": [instructor_score]}
    if children := trace_data.get()[instructor_trace_id.get()].children:
        last_child_trace_id = children[-1]
        trace_update_dict["configuration"] = trace_data.get()[last_child_trace_id].configuration
    trace_insert(trace_update_dict, instructor_trace_id.get())
    instructor_trace_id.set("")
    instructor_val_err_count.set(0)
    instructor_val_errs.set([])


class _RetryWrapper:
    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[type, Any],
        kwargs: Mapping[str, Any],
    ) -> Any:
        trace_id = gen_trace_id()
        instructor_trace_id.set(trace_id)
        try:
            inputs = kwargs.get("kwargs", {}).get("template_inputs", {})
            metadata = {}
            for key in ["max_retries", "response_model", "validation_context", "mode", "args"]:
                if kwargs.get(key):
                    metadata[key] = kwargs[key]
            trace_name = "instructor"
            if "response_model" in kwargs and kwargs["response_model"] and hasattr(kwargs["response_model"], "__name__"):
                trace_name = kwargs["response_model"].__name__

            def fn_transform_generator_outputs(items: List) -> str:
                try:
                    return json_dumps(items[-1])
                except Exception as e:
                    logger.warning(f"Failed to serialize generator output: {e}", exc_info=e)
                    return ""

            return trace(name=trace_name, overwrite_trace_id=trace_id, overwrite_inputs=inputs, metadata=metadata, fn_transform_generator_outputs=fn_transform_generator_outputs)(
                wrapped
            )(*args, **kwargs)
        except (InstructorRetryException, ValidationError, JSONDecodeError) as e:
            instructor_val_err_count.set(instructor_val_err_count.get() + 1)
            reasons = get_reasons(e)
            instructor_val_errs.set(instructor_val_errs.get() + reasons)

            report_instructor_validation_errors()
            current_log = trace_data.get()[trace_id]
            logger_update_record(UpdateLog(trace_id=trace_id, field_name_to_value_map={"scores": current_log.scores}, root_trace_id=current_log.root_trace_id))

            raise e


class _AttemptManagerExitWrapper:
    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[type, Any],
        kwargs: Mapping[str, Any],
    ) -> Any:
        if instructor_trace_id.get() is not None:
            if len(args) > 1 and args[1] is not None and isinstance(args[1], (InstructorRetryException, ValidationError, JSONDecodeError)):
                instructor_val_err_count.set(instructor_val_err_count.get() + 1)
                reasons = get_reasons(args[1])
                instructor_val_errs.set(instructor_val_errs.get() + reasons)
            else:
                report_instructor_validation_errors()
        return wrapped(*args, **kwargs)



from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, TypedDict, TypeVar, Union

import copy
import datetime
import functools
import json
import logging
import re
import uuid
from enum import Enum

from orjson import orjson

logger = logging.getLogger()


ID_TYPE = Union[uuid.UUID, str]
_MAX_DEPTH = 2


class RunTypeEnum(str, Enum):
    """(Deprecated) Enum for run types. Use string directly."""

    tool = "tool"
    chain = "chain"
    llm = "llm"
    retriever = "retriever"
    embedding = "embedding"
    prompt = "prompt"
    parser = "parser"


class RunLikeDict(TypedDict, total=False):
    """Run-like dictionary, for type-hinting."""

    name: str
    run_type: RunTypeEnum
    start_time: datetime
    inputs: Optional[dict]
    outputs: Optional[dict]
    end_time: Optional[datetime]
    extra: Optional[dict]
    error: Optional[str]
    serialized: Optional[dict]
    parent_run_id: Optional[uuid.UUID]
    manifest_id: Optional[uuid.UUID]
    events: Optional[List[dict]]
    tags: Optional[List[str]]
    inputs_s3_urls: Optional[dict]
    outputs_s3_urls: Optional[dict]
    id: Optional[uuid.UUID]
    session_id: Optional[uuid.UUID]
    session_name: Optional[str]
    reference_example_id: Optional[uuid.UUID]
    input_attachments: Optional[dict]
    output_attachments: Optional[dict]
    trace_id: uuid.UUID
    dotted_order: str


def _as_uuid(value: ID_TYPE, var: Optional[str] = None) -> uuid.UUID:
    try:
        return uuid.UUID(value) if not isinstance(value, uuid.UUID) else value
    except ValueError as e:
        var = var or "value"
        raise Exception(f"{var} must be a valid UUID or UUID string. Got {value}") from e


def _simple_default(obj: Any) -> Any:
    # Don't traverse into nested objects
    try:
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        if isinstance(obj, uuid.UUID):
            return str(obj)
        return json.loads(json.dumps(obj))
    except BaseException as e:
        logger.debug(f"Failed to serialize {type(obj)} to JSON: {e}")
        return repr(obj)


def _serialize_json(obj: Any, depth: int = 0, serialize_py: bool = True) -> Any:
    try:
        if depth >= _MAX_DEPTH:
            try:
                return orjson.loads(_dumps_json_single(obj))
            except BaseException:
                return repr(obj)
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        if isinstance(obj, (set, tuple)):
            return orjson.loads(_dumps_json_single(list(obj)))

        serialization_methods = [
            ("model_dump_json", True),  # Pydantic V2
            ("json", True),  # Pydantic V1
            ("to_json", False),  # dataclass_json
            ("model_dump", True),  # Pydantic V2 with non-serializable fields
            ("dict", False),  # Pydantic V1 with non-serializable fields
        ]
        for attr, exclude_none in serialization_methods:
            if hasattr(obj, attr) and callable(getattr(obj, attr)):
                try:
                    method = getattr(obj, attr)
                    json_str = method(exclude_none=exclude_none) if exclude_none else method()
                    if isinstance(json_str, str):
                        return json.loads(json_str)
                    return orjson.loads(_dumps_json(json_str, depth=depth + 1, serialize_py=serialize_py))
                except Exception as e:
                    logger.debug(f"Failed to serialize {type(obj)} to JSON: {e}")
                    pass
        if serialize_py:
            all_attrs = {}
            if hasattr(obj, "__slots__"):
                all_attrs.update({slot: getattr(obj, slot, None) for slot in obj.__slots__})
            if hasattr(obj, "__dict__"):
                all_attrs.update(vars(obj))
            if all_attrs:
                filtered = {k: v if v is not obj else repr(v) for k, v in all_attrs.items()}
                return orjson.loads(_dumps_json(filtered, depth=depth + 1, serialize_py=serialize_py))
        return repr(obj)
    except BaseException as e:
        logger.debug(f"Failed to serialize {type(obj)} to JSON: {e}")
        return repr(obj)


def _elide_surrogates(s: bytes) -> bytes:
    pattern = re.compile(rb"\\ud[89a-f][0-9a-f]{2}", re.IGNORECASE)
    result = pattern.sub(b"", s)
    return result


def _dumps_json_single(obj: Any, default: Optional[Callable[[Any], Any]] = None) -> bytes:
    try:
        return orjson.dumps(
            obj,
            default=default,
            option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SERIALIZE_DATACLASS | orjson.OPT_SERIALIZE_UUID | orjson.OPT_NON_STR_KEYS,
        )
    except TypeError as e:
        # Usually caused by UTF surrogate characters
        logger.debug(f"Orjson serialization failed: {repr(e)}. Falling back to json.")
        result = json.dumps(
            obj,
            default=_simple_default,
            ensure_ascii=True,
        ).encode("utf-8")
        try:
            result = orjson.dumps(orjson.loads(result.decode("utf-8", errors="surrogateescape")))
        except orjson.JSONDecodeError:
            result = _elide_surrogates(result)
        return result


def _dumps_json(obj: Any, depth: int = 0, serialize_py: bool = True) -> bytes:
    """Serialize an object to a JSON formatted string.
    Parameters
    ----------
    obj : Any
        The object to serialize.
    default : Callable[[Any], Any] or None, default=None
        The default function to use for serialization.
    Returns:
    -------
    str
        The JSON formatted string.
    """
    return _dumps_json_single(obj, functools.partial(_serialize_json, depth=depth, serialize_py=serialize_py))


T = TypeVar("T")


def _middle_copy(val: T, memo: Dict[int, Any], max_depth: int = 4, _depth: int = 0) -> T:
    cls = type(val)

    copier = getattr(cls, "__deepcopy__", None)
    if copier is not None:
        try:
            return copier(memo)
        except BaseException:
            pass
    if _depth >= max_depth:
        return val
    if isinstance(val, dict):
        return {_middle_copy(k, memo, max_depth, _depth + 1): _middle_copy(v, memo, max_depth, _depth + 1) for k, v in val.items()}  # type: ignore[return-value]
    if isinstance(val, list):
        return [_middle_copy(item, memo, max_depth, _depth + 1) for item in val]  # type: ignore[return-value]
    if isinstance(val, tuple):
        return tuple(_middle_copy(item, memo, max_depth, _depth + 1) for item in val)  # type: ignore[return-value]
    if isinstance(val, set):
        return {_middle_copy(item, memo, max_depth, _depth + 1) for item in val}  # type: ignore[return-value]

    return val


def deepish_copy(val: T) -> T:
    """Deep copy a value with a compromise for uncopyable objects.
    Args:
        val: The value to be deep copied.
    Returns:
        The deep copied value.
    """
    memo: Dict[int, Any] = {}
    try:
        return copy.deepcopy(val, memo)
    except BaseException as e:
        # Generators, locks, etc. cannot be copied
        # and raise a TypeError (mentioning pickling, since the dunder methods)
        # are re-used for copying. We'll try to do a compromise and copy
        # what we can
        logger.debug("Failed to deepcopy input: %s", repr(e))
        return _middle_copy(val, memo)
